#include "otabrowse5.h"
#pragma hdrstop

#include "sample.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------




// TWatcher.
// (1) A string property FarForm which contains the name of another form in the
//     project. As discussed in the header, we need a TDummyString class to
//     use it. Create this TDummyString object in the constructor, and free it
//     in the destructor. The Setter method does DummyString::Assign().
//     We RegisterPropertyEditor, and also RegisterPropertyWatcher so that
//     the value gets auto-updated whenever the form that it refers to is
//     renamed.
// (2) A property FarComponent which points to a component on another form.
//     (i.e. not just contains the name of another component, but is an
//     actual pointer to it). In the constructor it's initially set to null.
//     In the Setter method, we call FreeNotification(.) so we get told when
//     that other component gets deleted -- when it is, Notification is called,
//     and then we have to clear our pointer to it.
//
// What is the difference between having a pointer to a component on another
// form, and just having the name of it? Well, pointers are built into BCB
// and are more reliable. But it's impossible to have a pointer to another form,
// so in this case you have to have the name of it. Also, text names are more
// friendly to the user: if you have a pointer-property to another form, then
// it's not possible to view that other form as text. If you have a name-property
// with the name of a component on that other form, then you can view that
// other form as text. (note -- a devious user could choose to rename the
// component on that other form, so rendering that name invalid. That is the
// price we pay for having text-names rather than pointers. That's why I say
// that pointers are more reliable.)

__fastcall TWatcher::TWatcher(TComponent *Owner) : TComponent(Owner)
{ FFarForm=new TDummyString();
  FFarComponent=NULL;
}
__fastcall TWatcher::~TWatcher()
{ if (FFarForm!=NULL) delete FFarForm; FFarForm=NULL;
}
void __fastcall TWatcher::SetFarForm(TDummyString *ofc)
{ FFarForm->Assign(ofc);
}
void __fastcall TWatcher::SetFarComponent(TComponent *c)
{ if (FFarComponent==c) return;
  FFarComponent=c; if (FFarComponent!=NULL) FFarComponent->FreeNotification(this);
}
void __fastcall TWatcher::Notification(TComponent *com,TOperation op)
{ if (com==FFarComponent && op==opRemove) FFarComponent=NULL;
  TComponent::Notification(com,op);
}





// TFarComponentPropertyEditor -- this is a property editor for
// pointing to TComponents that can reside on other forms. It just
// gives a drop-down list of all components on all forms. See the
// documentation in toolsapiplus.h for the TFormScanner to see how
// you could customise this list, perhaps to display only components
// of a certain type.
class TFarComponentPropertyEditor : public TComponentProperty
{
public:
  TFormScanner *scanner;
  virtual __fastcall TFarComponentPropertyEditor(const _di_IFormDesigner ides,int count) : TComponentProperty(ides,count)
  { scanner = new TFormScanner();
  }
  virtual __fastcall ~TFarComponentPropertyEditor()
  { if (scanner!=NULL) delete scanner; scanner=NULL;
  }
  TPropertyAttributes __fastcall GetAttributes()
  { TPropertyAttributes pa; pa << paValueList << paSortList; return pa;
  }
  //
  void __fastcall GetValues(TGetStrProc Proc)
  { scanner->QuickScan(Designer,Proc,false,true,true);
  }
  AnsiString __fastcall GetValue()
  { TComponent *c=(TComponent*)GetOrdValue(); if (c==NULL) return "";
    AnsiString s=c->Name; if (c->Owner!=NULL) s=c->Owner->Name+"->"+s;
    return s;
  }
  void __fastcall SetValue(AnsiString NewValue)
  { scanner->QuickScan(Designer,NULL,false,true,true);
    if (scanner->OpenFormIfHidden(NewValue)) return;
    if (NewValue=="") {SetOrdValue(NULL);return;}
    int i=scanner->results->IndexOf(NewValue);
    if (i!=-1) SetOrdValue((int)scanner->results->Objects[i]);
  }
};




// TWatcherEditor. This component editor displays a popup-menu as soon
// as the component is dropped on the form. It then replaces this component
// with the one selected from the list.
class TWatcherEditor : public TSwappingComponentEditor
{
public:
  __fastcall TWatcherEditor(TComponent* Component, _di_IFormDesigner Designer) : TSwappingComponentEditor(Component,Designer) {}
protected:
  AnsiString __fastcall GetSwapText()
  { return "Swap...";
  }
  virtual void __fastcall ExecuteSwapMenu()
  { AddMenuItem("TWatcher","");
    AddMenuItem("-","");
    AddMenuItem("TButton","");
    AddMenuItem("TEdit","AnInitialName");
    int res=ExecuteMenu();
    if (res!=0) ExecuteSwap(res);
  }
};




// TDummyStringProperty -- because BCB won't let us have property-editors
// for strings. So we have to create a dummy class to property-edit.
class TDummyStringEditor : public TFormNameEditor
{
public:
  __fastcall TDummyStringEditor(const _di_IFormDesigner ides,int count) : TFormNameEditor(ides,count) {}
  //
  AnsiString __fastcall RetrieveValue()
  { TDummyString *ds=(TDummyString*)GetOrdValue();
    if (ds==NULL) return ""; else return ds->Val;
  }
  void __fastcall StoreValue(AnsiString NewValue)
  { TDummyString *ds=(TDummyString*)GetOrdValue();
    if (ds!=NULL) ds->Val=NewValue;
  }
};





namespace Sample
{ void __fastcall PACKAGE Register()
  { TComponentClass classes[] = {__classid(TWatcher)};
    RegisterComponents("OTABrowse",classes,(sizeof(classes)/sizeof(classes[0]))-1);

    // standard property editor
    RegisterPropertyEditor(__typeinfo(TComponent),__classid(TWatcher),"FarComponent",__classid(TFarComponentPropertyEditor));

    // RegisterPlonkVerb -- whenever a component of type TWatcher is dropped on
    // the form, its Swap... verb is automatically executed
    RegisterComponentEditor(__classid(TWatcher),__classid(TWatcherEditor));
    RegisterPlonkVerb("TWatcher","Swap...");

    // RegisterPropertyWatcher -- if this prop has the value 'Form1', and the user
    // renames it to Form2, the PropertyWatcher will automatically update the prop.
    RegisterPropertyEditor(__typeinfo(TDummyString),0,"",__classid(TDummyStringEditor));
    RegisterPropertyWatcher(__typeinfo(TDummyString),0,"",TDummyStringEditor::OnExists);
  }
}



