#include "otabrowse5.h"
#pragma hdrstop

#include "toolsapiplus.h"
#pragma package(smart_init)
//---------------------------------------------------------------------------
const AnsiString logfile="IDE";                // IDE or c:\\log.txt or *c:\\log.txt (last form only logs if file pre-existed)
const AnsiString title="OTA Browser";          // error dialog boxes use this
const AnsiString watcherid="WISCHIK.OTA5WATCH";// must be unique




//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// DEBUGGING FUNCTIONS
// db(s) - logs the message into logfile.
// dbe(s) - prints a message box as well as logging it
// dbee(s,e) - like dbe but also prints an exception's details
//   logfile can be 'IDE' (for messages in a persistent window in the IDE)
//   or 'c:\myfilelog.txt'
//   or '*c:\myfilelog.txt' (which won't log unless the file was already there).
// dmsg(TMessage), dptr(void*), di(int), dle()
//   all return a string format of the entity in question. dle is the last error.
//   If we're logging to a memo, that memo is owned by the application and
//   references no code or variables in this expert. Therefore the memo will
//   persist even after this expert has terminated.
// ref_add(ptr), ref_dec(ptr) - reference counting checks
//   Use these to check that all COM objects have been property freed.
//   In AddRef, add it; in ref_dec, decrement it.
// ref_check() - will check that all references have been freed
//   The reference checking is implemented with a fixed size hash table.
//   If you expect to have more than 50 or so instances live at any time,
//   you should change its size 'ref_max' to a larger prime 179,359,719,1439,2879
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
//
const int ref_max=89;
void*ref_ptr[ref_max]={0}; int ref_line[ref_max]={0}, ref_count[ref_max]={0}, ref_cur=0, ref_peak=0;
bool dbinited=false; bool dbfile=false; TMemo *dbmemo=NULL;
//
AnsiString __fastcall ExtractSafeName(TComponent *Owner,AnsiString orig,bool avoidduplicates)
{ AnsiString name="";
  for (int i=1; i<=orig.Length(); i++)
  { char c=orig[i]; bool ok=false;
    if (c>='a' && c<='z') ok=true;
    if (c>='A' && c<='Z') ok=true;
    if (c>='0' && c<='9') ok=true;
    if (c=='_') ok=true;
    if (ok) name=name+AnsiString(c);
  }
  if (Owner==NULL || !avoidduplicates) return name;
  AnsiString s=name;
  for (int i=1; Owner->FindComponent(s)!=NULL; i++) s=name+AnsiString(i);
  return s;
}
void __fastcall SafeSetName(TComponent *Owner,TComponent *c,AnsiString n)
{ c->Name=ExtractSafeName(Owner,n,true);
}
void __fastcall dbinit()
{ if (dbinited) return; dbinited=true; dbfile=false; dbmemo=NULL;
  if (logfile=="") return;
  if (logfile=="IDE")
  { if (Application->Terminated) return; // unlikely but conceivable!
    AnsiString fname=ExtractSafeName(Application,"form"+watcherid,false);
    TForm *dbform = dynamic_cast<TForm*>(Application->FindComponent(fname));
    if (dbform==NULL)
    { dbform=new TForm(Application,0); dbform->Name=fname; dbform->Caption=title; dbform->Width=400; dbform->Height=200; dbform->BorderStyle=bsSizeToolWin; dbform->FormStyle=fsStayOnTop;
      dbmemo=new TMemo(dbform); dbmemo->Name="lumemo"; dbmemo->Parent=dbform; dbmemo->Align=alClient; dbmemo->ScrollBars=ssVertical; dbmemo->Clear();
      dbform->Left=8; dbform->Top=Screen->Height-dbform->Height-40; dbform->Visible=true;
    }
    else
    { dbmemo = dynamic_cast<TMemo*>(dbform->FindComponent("lumemo"));
      if (dbform->Visible) dbmemo->Lines->Add("----------------");
      else dbmemo->Clear();
      dbform->Visible=true;
    }
    return;
  }
  AnsiString fn=logfile;
  if (fn[1]=='*')
  { fn=fn.SubString(2,fn.Length()-1);
    if (!FileExists(fn)) return;
  }
  FILE *f=fopen(fn.c_str(),"w+t");
  fputs("Starting log\n",f);
  fclose(f);
  dbfile=true;
}
void __fastcall db(AnsiString s)
{ dbinit();
  if (dbfile)
  { char *c=logfile.c_str(); if (*c=='*') c++;
    FILE *f=fopen(c,"a+t");
    fputs((s+"\n").c_str(),f);
    fclose(f);
  }
  else if (dbmemo!=NULL)
  { if (Application->Terminated) return; // otherwise it'd cause a crash,
    // as our expert attempted to recreate the TMemo's already-destroyed handle
    dbmemo->Lines->Add(s);
  }
}
void __fastcall dbe(AnsiString s)
{ db("*ERR* "+s);
  Application->MessageBox(s.c_str(),title.c_str(),MB_OK|MB_ICONEXCLAMATION);
}
void __fastcall dbee(AnsiString s,Exception &e)
{ dbe(s+" "+e.ClassName()+" "+e.Message);
}

AnsiString __fastcall dptr(void *ptr)
{ return "0x"+IntToHex((int)ptr,8);
}
AnsiString __fastcall di(int i)
{ return AnsiString(i);
}
AnsiString __fastcall dbool(bool b)
{ return b?"true":"false";
}
AnsiString __fastcall dle()
{ LPVOID lpMsgBuf;
  FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,NULL,GetLastError(),MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),(LPTSTR) &lpMsgBuf,0,NULL);
  AnsiString s=AnsiString((char*)lpMsgBuf);
  LocalFree( lpMsgBuf );
  while (s[s.Length()]=='\r' || s[s.Length()]=='\n') s=s.SubString(1,s.Length()-1);
  return s;
}
AnsiString __fastcall dwa(TWatchAction wa)
{ if (wa==waCreate) return "waCreate";
  else if (wa==waDestroy) return "waDestroy";
  else if (wa==waExists) return "waExists";
  else if (wa==waRename) return "waRename";
  else return "wa???";
}
AnsiString __fastcall dws(TWatchSubject ws)
{ if (ws==wsComponent) return "wsComponent";
  else if (ws==wsForm) return "wsForm";
  else if (ws==wsDebug) return "wsDebug";
  else return "ws???";
}
AnsiString __fastcall diid(REFIID riid)
{ if (riid==IID_IUnknown) return "IUnknown";
  else if (riid==IID_IBorlandIDEServices) return "IID_IBorlandIDEServices";
  else if (riid==IID_INTAComponent) return "IID_INTAComponent";
  else if (riid==IID_INTACustomDrawMessage) return "IID_INTACustomDrawMessage";
  else if (riid==IID_INTACustomMessage) return "IID_INTACustomMessage";
  else if (riid==IID_INTACustomMessage40) return "IID_INTACustomMessage40";
  else if (riid==IID_INTAFormEditor) return "IID_INTAFormEditor";
  else if (riid==IID_INTAServices) return "IID_INTAServices";
  else if (riid==IID_INTAServices40) return "IID_INTAServices40";
  else if (riid==IID_INTAToDoItem) return "IID_INTAToDoItem";
  else if (riid==IID_INTAUnknown) return "IID_INTAUnknown";
  else if (riid==IID_IOTAActionServices) return "IID_IOTAActionServices";
  else if (riid==IID_IOTAAddressBreakpoint) return "IID_IOTAAddressBreakpoint";
  else if (riid==IID_IOTABreakpoint) return "IID_IOTABreakpoint";
  else if (riid==IID_IOTABreakpoint40) return "IID_IOTABreakpoint40";
  else if (riid==IID_IOTABreakpointNotifier) return "IID_IOTABreakpointNotifier";
  else if (riid==IID_IOTABufferOptions) return "IID_IOTABufferOptions";
  else if (riid==IID_IOTAComponent) return "IID_IOTAComponent";
  else if (riid==IID_IOTACreator) return "IID_IOTACreator";
  else if (riid==IID_IOTADebuggerNotifier) return "IID_IOTADebuggerNotifier";
  else if (riid==IID_IOTADebuggerServices) return "IID_IOTADebuggerServices";
  else if (riid==IID_IOTAEditActions) return "IID_IOTAEditActions";
  else if (riid==IID_IOTAEditBlock) return "IID_IOTAEditBlock";
  else if (riid==IID_IOTAEditBuffer) return "IID_IOTAEditBuffer";
  else if (riid==IID_IOTAEditBufferIterator) return "IID_IOTAEditBufferIterator";
  else if (riid==IID_IOTAEditLineNotifier) return "IID_IOTAEditLineNotifier";
  else if (riid==IID_IOTAEditLineTracker) return "IID_IOTAEditLineTracker";
  else if (riid==IID_IOTAEditOptions) return "IID_IOTAEditOptions";
  else if (riid==IID_IOTAEditor) return "IID_IOTAEditor";
  else if (riid==IID_IOTAEditorNotifier) return "IID_IOTAEditorNotifier";
  else if (riid==IID_IOTAEditorServices) return "IID_IOTAEditorServices";
  else if (riid==IID_IOTAEditPosition) return "IID_IOTAEditPosition";
  else if (riid==IID_IOTAEditReader) return "IID_IOTAEditReader";
  else if (riid==IID_IOTAEditView) return "IID_IOTAEditView";
  else if (riid==IID_IOTAEditView40) return "IID_IOTAEditView40";
  else if (riid==IID_IOTAEditWriter) return "IID_IOTAEditWriter";
  else if (riid==IID_IOTAEnvironmentOptions) return "IID_IOTAEnvironmentOptions";
  else if (riid==IID_IOTAFile) return "IID_IOTAFile";
  else if (riid==IID_IOTAFileSystem) return "IID_IOTAFileSystem";
  else if (riid==IID_IOTAFormEditor) return "IID_IOTAFormEditor";
  else if (riid==IID_IOTAFormNotifier) return "IID_IOTAFormNotifier";
  else if (riid==IID_IOTAFormWizard) return "IID_IOTAFormWizard";
  else if (riid==IID_IOTAIDENotifier) return "IID_IOTAIDENotifier";
  else if (riid==IID_IOTAIDENotifier50) return "IID_IOTAIDENotifier50";
  else if (riid==IID_IOTAKeyBindingServices) return "IID_IOTAKeyBindingServices";
  else if (riid==IID_IOTAKeyboardBinding) return "IID_IOTAKeyboardBinding";
  else if (riid==IID_IOTAKeyboardServices) return "IID_IOTAKeyboardServices";
  else if (riid==IID_IOTAKeyboardDiagnostics) return "IID_IOTAKeyboardDiagnostics";
  else if (riid==IID_IOTAKeyContext) return "IID_IOTAKeyContext";
  else if (riid==IID_IOTAMenuWizard) return "IID_IOTAMenuWizard";
  else if (riid==IID_IOTAMessageServices) return "IID_IOTAMessageServices";
  else if (riid==IID_IOTAMessageServices40) return "IID_IOTAMessageServices40";
  else if (riid==IID_IOTAModule) return "IID_IOTAModule";
  else if (riid==IID_IOTAModuleCreator) return "IID_IOTAModuleCreator";
  else if (riid==IID_IOTAModuleInfo) return "IID_IOTAModuleInfo";
  else if (riid==IID_IOTAModuleNotifier) return "IID_IOTAModuleNotifier";
  else if (riid==IID_IOTAModuleServices) return "IID_IOTAModuleServices";
  else if (riid==IID_IOTANotifier) return "IID_IOTANotifier";
  else if (riid==IID_IOTAOptions) return "IID_IOTAOptions";
  else if (riid==IID_IOTAPackageServices) return "IID_IOTAPackageServices";
  else if (riid==IID_IOTAProcess) return "IID_IOTAProcess";
  else if (riid==IID_IOTAProcessModNotifier) return "IID_IOTAProcessModNotifier";
  else if (riid==IID_IOTAProcessModule) return "IID_IOTAProcessModule";
  else if (riid==IID_IOTAProcessNotifier) return "IID_IOTAProcessNotifier";
  else if (riid==IID_IOTAProject) return "IID_IOTAProject";
  else if (riid==IID_IOTAProjectBuilder) return "IID_IOTAProjectBuilder";
  else if (riid==IID_IOTAProjectBuilder40) return "IID_IOTAProjectBuilder40";
  else if (riid==IID_IOTAProjectCreator) return "IID_IOTAProjectCreator";
  else if (riid==IID_IOTAProjectGroup) return "IID_IOTAProjectGroup";
  else if (riid==IID_IOTAProjectGroupCreator) return "IID_IOTAProjectGroupCreator";
  else if (riid==IID_IOTAProjectOptions) return "IID_IOTAProjectOptions";
  else if (riid==IID_IOTAProjectOptions40) return "IID_IOTAProjectOptions40";
  else if (riid==IID_IOTAProjectResource) return "IID_IOTAProjectResource";
  else if (riid==IID_IOTAProjectWizard) return "IID_IOTAProjectWizard";
  else if (riid==IID_IOTARecord) return "IID_IOTARecord";
  else if (riid==IID_IOTAReplaceOptions) return "IID_IOTAReplaceOptions";
  else if (riid==IID_IOTARepositoryWizard) return "IID_IOTARepositoryWizard";
  else if (riid==IID_IOTAResourceEntry) return "IID_IOTAResourceEntry";
  else if (riid==IID_IOTASearchOptions) return "IID_IOTASearchOptions";
  else if (riid==IID_IOTAServices) return "IID_IOTAServices";
  else if (riid==IID_IOTASourceBreakpoint) return "IID_IOTASourceBreakpoint";
  else if (riid==IID_IOTASourceEditor) return "IID_IOTASourceEditor";
  else if (riid==IID_IOTASpeedSetting) return "IID_IOTASpeedSetting";
  else if (riid==IID_IOTAThread) return "IID_IOTAThread";
  else if (riid==IID_IOTAThreadNotifier) return "IID_IOTAThreadNotifier";
  else if (riid==IID_IOTAToDoManager) return "IID_IOTAToDoManager";
  else if (riid==IID_IOTAToDoServices) return "IID_IOTAToDoServices";
  else if (riid==IID_IOTATypeLibEditor) return "IID_IOTATypeLibEditor";
  else if (riid==IID_IOTATypeLibModule) return "IID_IOTATypeLibModule";
  else if (riid==IID_IOTAWizard) return "IID_IOTAWizard";
  else if (riid==IID_IOTAWizardServices) return "IID_IOTAWizardServices";
  else return "IID_???";
}
AnsiString __fastcall dmsgname(UINT msg)
{ switch (msg)
  { case WM_NULL: return "WM_NULL";
    case WM_CREATE: return "WM_CREATE";
    case WM_DESTROY: return "WM_DESTROY";
    case WM_MOVE: return "WM_MOVE";
    case WM_SIZE: return "WM_SIZE";
    case WM_ACTIVATE: return "WM_ACTIVATE";
    case WM_SETFOCUS: return "WM_SETFOCUS";
    case WM_KILLFOCUS: return "WM_KILLFOCUS";
    case WM_ENABLE: return "WM_ENABLE";
    case WM_SETREDRAW: return "WM_SETREDRAW";
    case WM_SETTEXT: return "WM_SETTEXT";
    case WM_GETTEXT: return "WM_GETTEXT";
    case WM_GETTEXTLENGTH: return "WM_GETTEXTLENGTH";
    case WM_PAINT: return "WM_PAINT";
    case WM_CLOSE: return "WM_CLOSE";
    case WM_QUERYENDSESSION: return "WM_QUERYENDSESSION";
    case WM_QUIT: return "WM_QUIT";
    case WM_QUERYOPEN: return "WM_QUERYOPEN";
    case WM_ERASEBKGND: return "WM_ERASEBKGND";
    case WM_SYSCOLORCHANGE: return "WM_SYSCOLORCHANGE";
    case WM_ENDSESSION: return "WM_ENDSESSION";
    case WM_SHOWWINDOW: return "WM_SHOWWINDOW";
    case WM_SETTINGCHANGE: return "WM_SETTINGCHANGE";
    case WM_DEVMODECHANGE: return "WM_DEVMODECHANGE";
    case WM_ACTIVATEAPP: return "WM_ACTIVATEAPP";
    case WM_FONTCHANGE: return "WM_FONTCHANGE";
    case WM_TIMECHANGE: return "WM_TIMECHANGE";
    case WM_CANCELMODE: return "WM_CANCELMODE";
    case WM_SETCURSOR: return "WM_SETCURSOR";
    case WM_MOUSEACTIVATE: return "WM_MOUSEACTIVATE";
    case WM_CHILDACTIVATE: return "WM_CHILDACTIVATE";
    case WM_QUEUESYNC: return "WM_QUEUESYNC";
    case WM_GETMINMAXINFO: return "WM_GETMINMAXINFO";
    case WM_PAINTICON: return "WM_PAINTICON";
    case WM_ICONERASEBKGND: return "WM_ICONERASEBKGND";
    case WM_NEXTDLGCTL: return "WM_NEXTDLGCTL";
    case WM_SPOOLERSTATUS: return "WM_SPOOLERSTATUS";
    case WM_DRAWITEM: return "WM_DRAWITEM";
    case WM_MEASUREITEM: return "WM_MEASUREITEM";
    case WM_DELETEITEM: return "WM_DELETEITEM";
    case WM_VKEYTOITEM: return "WM_VKEYTOITEM";
    case WM_CHARTOITEM: return "WM_CHARTOITEM";
    case WM_SETFONT: return "WM_SETFONT";
    case WM_GETFONT: return "WM_GETFONT";
    case WM_SETHOTKEY: return "WM_SETHOTKEY";
    case WM_GETHOTKEY: return "WM_GETHOTKEY";
    case WM_QUERYDRAGICON: return "WM_QUERYDRAGICON";
    case WM_COMPAREITEM: return "WM_COMPAREITEM";
    case WM_COMPACTING: return "WM_COMPACTING";
    case WM_COMMNOTIFY: return "WM_COMMNOTIFY";
    case WM_WINDOWPOSCHANGING: return "WM_WINDOWPOSCHANGING";
    case WM_WINDOWPOSCHANGED: return "WM_WINDOWPOSCHANGED";
    case WM_POWER: return "WM_POWER";
    case WM_COPYDATA: return "WM_COPYDATA";
    case WM_CANCELJOURNAL: return "WM_CANCELJOURNAL";
    case WM_NOTIFY: return "WM_NOTIFY";
    case WM_INPUTLANGCHANGEREQUEST: return "WM_INPUTLANGCHANGEREQUEST";
    case WM_INPUTLANGCHANGE: return "WM_INPUTLANGCHANGE";
    case WM_TCARD: return "WM_TCARD";
    case WM_HELP: return "WM_HELP";
    case WM_USERCHANGED: return "WM_USERCHANGED";
    case WM_NOTIFYFORMAT: return "WM_NOTIFYFORMAT";
    case WM_CONTEXTMENU: return "WM_CONTEXTMENU";
    case WM_STYLECHANGING: return "WM_STYLECHANGING";
    case WM_STYLECHANGED: return "WM_STYLECHANGED";
    case WM_DISPLAYCHANGE: return "WM_DISPLAYCHANGE";
    case WM_GETICON: return "WM_GETICON";
    case WM_SETICON: return "WM_SETICON";
    case WM_NCCREATE: return "WM_NCCREATE";
    case WM_NCDESTROY: return "WM_NCDESTROY";
    case WM_NCCALCSIZE: return "WM_NCCALCSIZE";
    case WM_NCHITTEST: return "WM_NCHITTEST";
    case WM_NCPAINT: return "WM_NCPAINT";
    case WM_NCACTIVATE: return "WM_NCACTIVATE";
    case WM_GETDLGCODE: return "WM_GETDLGCODE";
    case WM_NCMOUSEMOVE: return "WM_NCMOUSEMOVE";
    case WM_NCLBUTTONDOWN: return "WM_NCLBUTTONDOWN";
    case WM_NCLBUTTONUP: return "WM_NCLBUTTONUP";
    case WM_NCLBUTTONDBLCLK: return "WM_NCLBUTTONDBLCLK";
    case WM_NCRBUTTONDOWN: return "WM_NCRBUTTONDOWN";
    case WM_NCRBUTTONUP: return "WM_NCRBUTTONUP";
    case WM_NCRBUTTONDBLCLK: return "WM_NCRBUTTONDBLCLK";
    case WM_NCMBUTTONDOWN: return "WM_NCMBUTTONDOWN";
    case WM_NCMBUTTONUP: return "WM_NCMBUTTONUP";
    case WM_NCMBUTTONDBLCLK: return "WM_NCMBUTTONDBLCLK";
    case WM_KEYDOWN: return "WM_KEYDOWN";
    case WM_KEYUP: return "WM_KEYUP";
    case WM_CHAR: return "WM_CHAR";
    case WM_DEADCHAR: return "WM_DEADCHAR";
    case WM_SYSKEYDOWN: return "WM_SYSKEYDOWN";
    case WM_SYSKEYUP: return "WM_SYSKEYUP";
    case WM_SYSCHAR: return "WM_SYSCHAR";
    case WM_SYSDEADCHAR: return "WM_SYSDEADCHAR";
    case WM_IME_STARTCOMPOSITION: return "WM_IME_STARTCOMPOSITION";
    case WM_IME_ENDCOMPOSITION: return "WM_IME_ENDCOMPOSITION";
    case WM_IME_COMPOSITION: return "WM_IME_COMPOSITION";
    case WM_INITDIALOG: return "WM_INITDIALOG";
    case WM_COMMAND: return "WM_COMMAND";
    case WM_SYSCOMMAND: return "WM_SYSCOMMAND";
    case WM_TIMER: return "WM_TIMER";
    case WM_HSCROLL: return "WM_HSCROLL";
    case WM_VSCROLL: return "WM_VSCROLL";
    case WM_INITMENU: return "WM_INITMENU";
    case WM_INITMENUPOPUP: return "WM_INITMENUPOPUP";
    case WM_MENUSELECT: return "WM_MENUSELECT";
    case WM_MENUCHAR: return "WM_MENUCHAR";
    case WM_ENTERIDLE: return "WM_ENTERIDLE";
    case WM_CTLCOLORMSGBOX: return "WM_CTLCOLORMSGBOX";
    case WM_CTLCOLOREDIT: return "WM_CTLCOLOREDIT";
    case WM_CTLCOLORLISTBOX: return "WM_CTLCOLORLISTBOX";
    case WM_CTLCOLORBTN: return "WM_CTLCOLORBTN";
    case WM_CTLCOLORDLG: return "WM_CTLCOLORDLG";
    case WM_CTLCOLORSCROLLBAR: return "WM_CTLCOLORSCROLLBAR";
    case WM_CTLCOLORSTATIC: return "WM_CTLCOLORSTATIC";
    case WM_MOUSEMOVE: return "WM_MOUSEMOVE";
    case WM_LBUTTONDOWN: return "WM_LBUTTONDOWN";
    case WM_LBUTTONUP: return "WM_LBUTTONUP";
    case WM_LBUTTONDBLCLK: return "WM_LBUTTONDBLCLK";
    case WM_RBUTTONDOWN: return "WM_RBUTTONDOWN";
    case WM_RBUTTONUP: return "WM_RBUTTONUP";
    case WM_RBUTTONDBLCLK: return "WM_RBUTTONDBLCLK";
    case WM_MBUTTONDOWN: return "WM_MBUTTONDOWN";
    case WM_MBUTTONUP: return "WM_MBUTTONUP";
    case WM_MBUTTONDBLCLK: return "WM_MBUTTONDBLCLK";
    case WM_MOUSEWHEEL: return "WM_MOUSEWHEEL";
    case WM_PARENTNOTIFY: return "WM_PARENTNOTIFY";
    case WM_ENTERMENULOOP: return "WM_ENTERMENULOOP";
    case WM_EXITMENULOOP: return "WM_EXITMENULOOP";
    case WM_NEXTMENU: return "WM_NEXTMENU";
    case WM_SIZING: return "WM_SIZING";
    case WM_CAPTURECHANGED: return "WM_CAPTURECHANGED";
    case WM_MOVING: return "WM_MOVING";
    case WM_POWERBROADCAST: return "WM_POWERBROADCAST";
    case WM_DEVICECHANGE: return "WM_DEVICECHANGE";
    case WM_IME_SETCONTEXT: return "WM_IME_SETCONTEXT";
    case WM_IME_NOTIFY: return "WM_IME_NOTIFY";
    case WM_IME_CONTROL: return "WM_IME_CONTROL";
    case WM_IME_COMPOSITIONFULL: return "WM_IME_COMPOSITIONFULL";
    case WM_IME_SELECT: return "WM_IME_SELECT";
    case WM_IME_CHAR: return "WM_IME_CHAR";
    case WM_IME_KEYDOWN: return "WM_IME_KEYDOWN";
    case WM_IME_KEYUP: return "WM_IME_KEYUP";
    case WM_MDICREATE: return "WM_MDICREATE";
    case WM_MDIDESTROY: return "WM_MDIDESTROY";
    case WM_MDIACTIVATE: return "WM_MDIACTIVATE";
    case WM_MDIRESTORE: return "WM_MDIRESTORE";
    case WM_MDINEXT: return "WM_MDINEXT";
    case WM_MDIMAXIMIZE: return "WM_MDIMAXIMIZE";
    case WM_MDITILE: return "WM_MDITILE";
    case WM_MDICASCADE: return "WM_MDICASCADE";
    case WM_MDIICONARRANGE: return "WM_MDIICONARRANGE";
    case WM_MDIGETACTIVE: return "WM_MDIGETACTIVE";
    case WM_MDISETMENU: return "WM_MDISETMENU";
    case WM_ENTERSIZEMOVE: return "WM_ENTERSIZEMOVE";
    case WM_EXITSIZEMOVE: return "WM_EXITSIZEMOVE";
    case WM_DROPFILES: return "WM_DROPFILES";
    case WM_MDIREFRESHMENU: return "WM_MDIREFRESHMENU";
    case WM_MOUSEHOVER: return "WM_MOUSEHOVER";
    case WM_MOUSELEAVE: return "WM_MOUSELEAVE";
    case WM_CUT: return "WM_CUT";
    case WM_COPY: return "WM_COPY";
    case WM_PASTE: return "WM_PASTE";
    case WM_CLEAR: return "WM_CLEAR";
    case WM_UNDO: return "WM_UNDO";
    case WM_RENDERFORMAT: return "WM_RENDERFORMAT";
    case WM_RENDERALLFORMATS: return "WM_RENDERALLFORMATS";
    case WM_DESTROYCLIPBOARD: return "WM_DESTROYCLIPBOARD";
    case WM_DRAWCLIPBOARD: return "WM_DRAWCLIPBOARD";
    case WM_PAINTCLIPBOARD: return "WM_PAINTCLIPBOARD";
    case WM_VSCROLLCLIPBOARD: return "WM_VSCROLLCLIPBOARD";
    case WM_SIZECLIPBOARD: return "WM_SIZECLIPBOARD";
    case WM_ASKCBFORMATNAME: return "WM_ASKCBFORMATNAME";
    case WM_CHANGECBCHAIN: return "WM_CHANGECBCHAIN";
    case WM_HSCROLLCLIPBOARD: return "WM_HSCROLLCLIPBOARD";
    case WM_QUERYNEWPALETTE: return "WM_QUERYNEWPALETTE";
    case WM_PALETTEISCHANGING: return "WM_PALETTEISCHANGING";
    case WM_PALETTECHANGED: return "WM_PALETTECHANGED";
    case WM_HOTKEY: return "WM_HOTKEY";
    case WM_PRINT: return "WM_PRINT";
    case WM_PRINTCLIENT: return "WM_PRINTCLIENT";
    case WM_HANDHELDFIRST: return "WM_HANDHELDFIRST";
    case WM_HANDHELDLAST: return "WM_HANDHELDLAST";
    case WM_AFXFIRST: return "WM_AFXFIRST";
    case WM_AFXLAST: return "WM_AFXLAST";
    case WM_PENWINFIRST: return "WM_PENWINFIRST";
    case WM_PENWINLAST: return "WM_PENWINLAST";
    case CM_ACTIVATE: return "CM_ACTIVATE";
    case CM_DEACTIVATE: return "CM_DEACTIVATE";
    case CM_GOTFOCUS: return "CM_GOTFOCUS";
    case CM_LOSTFOCUS: return "CM_LOSTFOCUS";
    case CM_CANCELMODE: return "CM_CANCELMODE";
    case CM_DIALOGKEY: return "CM_DIALOGKEY";
    case CM_DIALOGCHAR: return "CM_DIALOGCHAR";
    case CM_FOCUSCHANGED: return "CM_FOCUSCHANGED";
    case CM_PARENTFONTCHANGED: return "CM_PARENTFONTCHANGED";
    case CM_PARENTCOLORCHANGED: return "CM_PARENTCOLORCHANGED";
    case CM_HITTEST: return "CM_HITTEST";
    case CM_VISIBLECHANGED: return "CM_VISIBLECHANGED";
    case CM_ENABLEDCHANGED: return "CM_ENABLEDCHANGED";
    case CM_COLORCHANGED: return "CM_COLORCHANGED";
    case CM_FONTCHANGED: return "CM_FONTCHANGED";
    case CM_CURSORCHANGED: return "CM_CURSORCHANGED";
    case CM_CTL3DCHANGED: return "CM_CTL3DCHANGED";
    case CM_PARENTCTL3DCHANGED: return "CM_PARENTCTL3DCHANGED";
    case CM_TEXTCHANGED: return "CM_TEXTCHANGED";
    case CM_MOUSEENTER: return "CM_MOUSEENTER";
    case CM_MOUSELEAVE: return "CM_MOUSELEAVE";
    case CM_MENUCHANGED: return "CM_MENUCHANGED";
    case CM_APPKEYDOWN: return "CM_APPKEYDOWN";
    case CM_APPSYSCOMMAND: return "CM_APPSYSCOMMAND";
    case CM_BUTTONPRESSED: return "CM_BUTTONPRESSED";
    case CM_SHOWINGCHANGED: return "CM_SHOWINGCHANGED";
    case CM_ENTER: return "CM_ENTER";
    case CM_EXIT: return "CM_EXIT";
    case CM_DESIGNHITTEST: return "CM_DESIGNHITTEST";
    case CM_ICONCHANGED: return "CM_ICONCHANGED";
    case CM_WANTSPECIALKEY: return "CM_WANTSPECIALKEY";
    case CM_INVOKEHELP: return "CM_INVOKEHELP";
    case CM_WINDOWHOOK: return "CM_WINDOWHOOK";
    case CM_RELEASE: return "CM_RELEASE";
    case CM_SHOWHINTCHANGED: return "CM_SHOWHINTCHANGED";
    case CM_PARENTSHOWHINTCHANGED: return "CM_PARENTSHOWHINTCHANGED";
    case CM_SYSCOLORCHANGE: return "CM_SYSCOLORCHANGE";
    case CM_WININICHANGE: return "CM_WININICHANGE";
    case CM_FONTCHANGE: return "CM_FONTCHANGE";
    case CM_TIMECHANGE: return "CM_TIMECHANGE";
    case CM_TABSTOPCHANGED: return "CM_TABSTOPCHANGED";
    case CM_UIACTIVATE: return "CM_UIACTIVATE";
    case CM_UIDEACTIVATE: return "CM_UIDEACTIVATE";
    case CM_DOCWINDOWACTIVATE: return "CM_DOCWINDOWACTIVATE";
    case CM_CONTROLLISTCHANGE: return "CM_CONTROLLISTCHANGE";
    case CM_GETDATALINK: return "CM_GETDATALINK";
    case CM_CHILDKEY: return "CM_CHILDKEY";
    case CM_DRAG: return "CM_DRAG";
    case CM_HINTSHOW: return "CM_HINTSHOW";
    case CM_DIALOGHANDLE: return "CM_DIALOGHANDLE";
    case CM_ISTOOLCONTROL: return "CM_ISTOOLCONTROL";
    case CM_RECREATEWND: return "CM_RECREATEWND";
    case CM_INVALIDATE: return "CM_INVALIDATE";
    case CM_SYSFONTCHANGED: return "CM_SYSFONTCHANGED";
    case CM_CONTROLCHANGE: return "CM_CONTROLCHANGE";
    case CM_CHANGED: return "CM_CHANGED";
    case CN_CHARTOITEM: return "CN_CHARTOITEM";
    case CN_COMMAND: return "CN_COMMAND";
    case CN_COMPAREITEM: return "CN_COMPAREITEM";
    case CN_CTLCOLORBTN: return "CN_CTLCOLORBTN";
    case CN_CTLCOLORDLG: return "CN_CTLCOLORDLG";
    case CN_CTLCOLOREDIT: return "CN_CTLCOLOREDIT";
    case CN_CTLCOLORLISTBOX: return "CN_CTLCOLORLISTBOX";
    case CN_CTLCOLORMSGBOX: return "CN_CTLCOLORMSGBOX";
    case CN_CTLCOLORSCROLLBAR: return "CN_CTLCOLORSCROLLBAR";
    case CN_CTLCOLORSTATIC: return "CN_CTLCOLORSTATIC";
    case CN_DELETEITEM: return "CN_DELETEITEM";
    case CN_DRAWITEM: return "CN_DRAWITEM";
    case CN_HSCROLL: return "CN_HSCROLL";
    case CN_MEASUREITEM: return "CN_MEASUREITEM";
    case CN_PARENTNOTIFY: return "CN_PARENTNOTIFY";
    case CN_VKEYTOITEM: return "CN_VKEYTOITEM";
    case CN_VSCROLL: return "CN_VSCROLL";
    case CN_KEYDOWN: return "CN_KEYDOWN";
    case CN_KEYUP: return "CN_KEYUP";
    case CN_CHAR: return "CN_CHAR";
    case CN_SYSKEYDOWN: return "CN_SYSKEYDOWN";
    case CN_SYSCHAR: return "CN_SYSCHAR";
    case CN_NOTIFY: return "CN_NOTIFY";
    default: return "WM_????";
  }
}
AnsiString __fastcall dmsg(TMessage &Msg)
{ return dmsgname(Msg.Msg)+" wparam="+Msg.WParam+" lparam="+Msg.LParam;
}

// reference counting is done with a fixed-size has table.
// ref_ptrs[] is an array of pointers.
// We'd like to store pointer p in p%ref_max if possible.
// But if that slot's already taken it, we'll use the next available slot.
// ref_line[p%ref_max] says how far ahead the furthest one has had to move
void __fastcall ref_add(void *ptr)
{ // look in the hash-table to see if this fellow already exists
  int sugg=((int)ptr)%ref_max;
  int guess=sugg;
  int len=ref_line[sugg];
  for (int i=0; i<len; i++)
  { if (ref_ptr[guess]==ptr) {ref_count[guess]++; return;}
    guess++; if (guess==ref_max) guess=0;
  }
  // it doesn't. So, new entry time.
  sugg=((int)ptr)%ref_max;
  guess=sugg;
  for (int i=0; i<ref_max; i++)
  { if (ref_ptr[guess]==NULL)
    { if (i+1>ref_line[sugg]) ref_line[sugg]=i+1;
      ref_cur++; if (ref_cur>ref_peak) ref_peak=ref_cur;
      ref_ptr[guess]=ptr; ref_count[guess]=1; return;
    }
    guess++; if (guess==ref_max) guess=0;
  }
  dbe("Out of room in ref-counting table.");
}
void __fastcall ref_dec(void *ptr)
{ int sugg=((int)ptr)%ref_max;
  int guess=sugg;
  int len=ref_line[sugg];
  for (int i=0; i<len; i++)
  { if (ref_ptr[guess]==ptr)
    { ref_count[guess] --;
      if (ref_count[guess]>0) return;
      ref_cur--;
      ref_ptr[guess]=0;
      // we found the ptr here. Time to dec it, and maybe reduce the line
      if (i!=len-1) return;
      for (int j=i-1; j>=0; --j)
      { void*usedptr = ref_ptr[(sugg+j)%ref_max];
        int  usedhash= ((int)usedptr)%ref_max;
        if (usedhash==sugg) {ref_line[sugg]=j+1;return;}
      }
      ref_line[sugg]=0; return;
    }
    guess++; if (guess==ref_max) guess=0;
  }
  dbe("Trying to decrement a non-existent pointer "+dptr(ptr));
}
void __fastcall ref_check()
{ bool anyleft=false;
  for (int i=0; i<ref_max; i++)
  { if (ref_ptr[i] != NULL && ref_count[i]!=0)
    { db("Object at "+dptr(ref_ptr[i])+" left with "+di(ref_count[i])+" counts");
      anyleft=true;
    }
  }
  if (anyleft) dbe("Not all objects were released.");
}





// ExtractWatchEventSubject -- a yucky syntax for extracting the subject of a closure
TComponent* __fastcall ExtractWatchEventSubject(TWatchEvent we) {if (we==NULL) return NULL; else return (TComponent*)(((void**)&we)[1]);}











//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// IOTA NAVIGATION FUNCTIONS
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// Helpful ways to get from one IOTA to another.
//
_di_IDesigner __fastcall GetDesignerFromFormEditor(_di_IOTAFormEditor fe)
{ if (fe==NULL) {db("GetDesignerFromFormEditor: given null fe");return NULL;}
  _di_INTAFormEditor nfe=NULL; HRESULT res=fe->QueryInterface(__uuidof(INTAFormEditor),(void**)&nfe);
  if (FAILED(res)) {db("GetDesignerFromFormEditor: failed to get nfe");return NULL;}
  _di_IDesigner ides=nfe->GetFormDesigner();
  if (ides==NULL) db("GetDesignerFromFormEditor: ides was null");
  return ides;
}

_di_IOTAModuleServices __fastcall GetModuleServices()
{ if (BorlandIDEServices==NULL) {db("GetModuleServices: BorlandIDEServices are null");return NULL;}
  _di_IOTAModuleServices ms=NULL; HRESULT res=BorlandIDEServices->QueryInterface(__uuidof(IOTAModuleServices),(void**)&ms);
  if (FAILED(res)) db("GetModuleServices: failed");
  return ms;
}

_di_IOTAModule __fastcall GetModuleFromFileName(AnsiString un)
{ return GetModuleServices()->FindModule(un);
}


_di_IOTAFormEditor __fastcall GetFormEditorForModule(_di_IOTAModule m)
{ if (m==NULL) return NULL;
  _di_IOTAFormEditor fe=NULL;
  for (int i=0; i<m->GetModuleFileCount() && fe==NULL; i++)
  { _di_IOTAEditor e=NULL;
    try {e=m->GetModuleFileEditor(i);} catch (...) {};
    if (e!=NULL) fe=e; // that does an implicit dynamic_cast
  }
  return fe;
}



_di_IOTAProject __fastcall GetModuleAsProject(_di_IOTAModule m)
{ _di_IOTAProject p=NULL;
  HRESULT res=m->QueryInterface(__uuidof(IOTAProject),(void**)&p);
  if (FAILED(res)) return NULL; else return p;
}


// GetProjectFromDesigner -- we will first enumerate all files in the designer.
// We will then go through all open projects, until we find one that itself
// contains all those files. There is a pathalogical possibility of error,
// in the case where one project is strictly greater than another. But hey, who cares?!
class TGetProjectFromDesigner : public TObject
{ public:
  TStringList *fns;
  __fastcall TGetProjectFromDesigner() : TObject() {fns=new TStringList();}
  __fastcall ~TGetProjectFromDesigner() {if (fns!=NULL) delete fns; fns=NULL;}
  void __fastcall GetModuleProc(const AnsiString FileName, const AnsiString UnitName,const AnsiString FormName, const AnsiString DesignClass, TStrings* CoClasses) {fns->Add(FileName);}
};
_di_IOTAProject __fastcall GetProjectFromDesigner(_di_IFormDesigner ides)
{ TGetProjectFromDesigner *gpfd=new TGetProjectFromDesigner();
  ides->GetProjectModules(gpfd->GetModuleProc);
  _di_IOTAModuleServices ms=GetModuleServices();
  _di_IOTAProject res=NULL;
  for (int pi=0; res==NULL && pi<ms->ModuleCount; pi++)
  { _di_IOTAProject p=GetModuleAsProject(ms->Modules[pi]);
    if (p!=NULL)
    { TStringList *reqs=new TStringList();
      for (int i=0; i<gpfd->fns->Count; i++) reqs->Add(gpfd->fns->Strings[i]);
      for (int mi=0; mi<p->GetModuleCount(); mi++)
      { _di_IOTAModuleInfo modi = p->GetModule(mi);
        int i=reqs->IndexOf(modi->GetFileName());
        if (i!=-1) reqs->Delete(i);
      }
      AnsiString s=""; for (int i=0; i<reqs->Count; i++) s=s+reqs->Strings[i]+" ";
      if (reqs->Count==0) res=p;
      delete reqs;
    }
  }
  delete gpfd;
  if (res==NULL) {dbe("GetProjectFromDesigner. Project not found.");}
  return res;
}



_di_IOTAFormEditor GetFormEditorEx(AnsiString fo,AnsiString un)
{ _di_IOTAModuleServices ms=NULL; HRESULT res=BorlandIDEServices->QueryInterface(__uuidof(IOTAModuleServices),(void**)&ms); if (FAILED(res)) return NULL;
  _di_IOTAModule m=ms->FindFormModule(fo);
  if (m==NULL) // maybe it just wasn't opened yet. In which case we'll try to open it
  { // We'll go through all opened project modules (starting with the current project)
    // and inside each project we'll go through all its constituent moduleinfos.
    for (int i=-1; i<ms->ModuleCount && m==NULL; i++)
    { _di_IOTAProject p=NULL;
      if (i==-1) p=GetCurrentProject();
      else {_di_IOTAModule m=ms->Modules[i]; m->QueryInterface(__uuidof(IOTAProject),(void**)&p);}
      if (p!=NULL)
      { for (int j=0; j<p->GetModuleCount() && m==NULL; j++)
        { _di_IOTAModuleInfo mi=p->GetModule(j);
          int mt=mi->GetModuleType(); AnsiString mun=mi->GetFileName(); AnsiString mfo=mi->GetFormName();
          if (mt==0 && mun==un && mfo==fo) m=mi->OpenModule();
        }
      }
    }
  }
  return GetFormEditorForModule(m);
}

// GETPROJECTCONTAININGCOMPONENT
// Tricky one this! It's needed because maybe a component initialises itself
// at designtime, and it has a pointer to another form in the project. It needs
// notification for when a form gets renamed in the same project. Therefore
// it needs to know what project it's in.
//   You might think to scan through all modules' form-editors, and use
// GetComponentHandle to see if the component is in there. But no! Because
// when you open a form, the component gets constructed before the FormEditor
// has been created. Therefore, if you tried to GetProjectContainingComponent
// in the component's constructor, it wouldn't work.
//   Instead we go up the TComponent's hierarchy to find the name of the
// form that ultimately owns it. Then search the ModuleInfos for a module with
// this form name.
_di_IOTAModule __fastcall GetModuleContainingComponent(TComponent *com)
{ if (com==NULL) return NULL; while (com->Owner!=NULL) com=com->Owner;
  AnsiString form=com->Name;
  _di_IOTAModuleServices ms=NULL; HRESULT res=BorlandIDEServices->QueryInterface(__uuidof(IOTAModuleServices),(void**)&ms); if (FAILED(res)) return NULL;
  for (int ni=0; ni<ms->GetModuleCount(); ni++)
  { _di_IOTAModule m=ms->GetModule(ni);
    _di_IOTAProject p=NULL; res=m->QueryInterface(__uuidof(IOTAProject),(void**)&p);
    if (SUCCEEDED(res))
    { for (int i=0; i<p->GetModuleCount(); i++)
      { _di_IOTAModuleInfo mi=p->GetModule(i);
        if (mi->GetFormName()==form) return mi->OpenModule();
      }
    }
  }
  return NULL;
}



// GETPROJECTFROMFILENAME
// Given c:\proj1.bpk we find its project
_di_IOTAProject __fastcall GetProjectFromFileName(AnsiString fn)
{ _di_IOTAModuleServices ms=BorlandIDEServices;
  _di_IOTAModule m=ms->FindModule(fn); if (m==NULL) return NULL;
  _di_IOTAProject p=m; // that does an implicit dynamic_cast
  return p;
}






// GetModuleInfoFromFileName - only works if the module is in a project, any project.
_di_IOTAModuleInfo __fastcall GetModuleInfoFromFileName(AnsiString fn)
{ _di_IOTAModuleServices ms = BorlandIDEServices;
  for (int i=0; i<ms->ModuleCount; i++)
  { _di_IOTAProject p = ms->Modules[i]; // does an implicit dynamic_cast
    if (p!=NULL)
    { for (int j=0; j<p->GetModuleCount(); j++)
      { if (fn.LowerCase() == p->GetModule(j)->GetFileName().LowerCase())
        { return p->GetModule(j);
        }
      }
    }
  }
  return NULL;
}






// GETCURRENTPROJECT
// Either there is no project group open, and just one project
// Or there is a project group open with its idea of the current project
_di_IOTAProject __fastcall GetCurrentProject()
{ _di_IOTAModuleServices ms=NULL;
  HRESULT res=BorlandIDEServices->QueryInterface(__uuidof(IOTAModuleServices),(void**)&ms);
  if (FAILED(res)) return NULL;
  int foundproject=-1;
  for (int i=0; i<ms->ModuleCount; i++)
  { _di_IOTAModule m=ms->Modules[i];
    _di_IOTAProjectGroup pg=NULL;
    _di_IOTAProject p=NULL;
    res=m->QueryInterface(__uuidof(IOTAProjectGroup),(void**)&pg);
    if (SUCCEEDED(res)) return pg->GetActiveProject();
    res=m->QueryInterface(__uuidof(IOTAProject),(void**)&p);
    if (SUCCEEDED(res)) foundproject=i;
  }
  if (foundproject==-1) return NULL;
  _di_IOTAProject p=NULL;
  ms->Modules[foundproject]->QueryInterface(__uuidof(IOTAProject),(void**)&p);
  return p;
}
_di_IOTAProjectResource __fastcall GetProjectResources(_di_IOTAProject p)
{ if (p==NULL) p=GetCurrentProject();
  if (p==NULL) return NULL;
  int num=p->GetModuleFileCount();
  for (int i=0; i<num; i++)
  { _di_IOTAEditor e=p->GetModuleFileEditor(i);
    _di_IOTAProjectResource pr;
    HRESULT res=e->QueryInterface(__uuidof(IOTAProjectResource),(void**)&pr);
    if (SUCCEEDED(res)) return pr;
  }
  return NULL;
}

bool __fastcall ExtractLastNumber(AnsiString *os,int *on)
{ AnsiString s=*os; if (s=="") {*on=0;return false;}
  int lastrealchar=s.Length();
  while (s[lastrealchar]>='0' && s[lastrealchar]<='9' && lastrealchar>1) lastrealchar--;
  AnsiString snum=s.SubString(lastrealchar+1,s.Length()-lastrealchar);
  if (snum=="") *on=0; else *on=snum.ToInt();
  *os=s.SubString(1,lastrealchar);
  return (snum!="");
}

void __fastcall EnsureUniqueModuleName(AnsiString *ounit,AnsiString *oform,AnsiString *odir)
{ AnsiString unit="",form="",fn="",dir="";
  // First, just in order to have something at least, we ask for any name
  // It will give the answer "unit1.cpp prefix1 c:\unit1.cpp"
  // or maybe "unit2.cpp prefix2 c:\unit2.cpp" if the name Prefix1 had
  // already been taken.
  AnsiString prefix=""; if (oform!=NULL) prefix=*oform;
  _di_IOTAModuleServices ms=GetModuleServices();
  ms->GetNewModuleAndClassName(prefix,unit,form,fn);
  dir=ExtractFilePath(fn);
  if (ounit!=NULL && *ounit!="") unit=*ounit;
  if (oform!=NULL && *oform!="") form=*oform;
  if (odir!=NULL && *odir!="") fn=*odir+unit;
  //
  // At this stage, we have suggested names all round.
  // If they're already numbered, then subsequently we'll be able to increment.
  // If they're not already numbered, then subsequently we'll have to invent numbers
  int number, i; bool isalreadytaken=true;
  AnsiString unbase=unit; bool isunitnumbered=ExtractLastNumber(&unbase,&number);
  AnsiString fobase=form; bool isformnumbered=ExtractLastNumber(&fobase,&i); if (i>number) number=i;
  while (isalreadytaken)    
  { isalreadytaken=false;
    unit=unbase; form=fobase;
    if (isunitnumbered) unit=unit+AnsiString(number);
    if (isformnumbered) form=form+AnsiString(number);
    for (int mi=0; mi<ms->ModuleCount; mi++)
    { _di_IOTAProject p=GetModuleAsProject(ms->Modules[mi]);
      for (int pi=0; p!=NULL && pi<p->GetModuleCount(); pi++)
      { _di_IOTAModuleInfo mi=p->GetModule(pi);
        if (mi->GetName()==unit) isalreadytaken=true;
        if (mi->GetFileName()==fn) isalreadytaken=true;
        if (form!="" && mi->GetFormName()==form) {isalreadytaken=true;isformnumbered=true;}
      }
    }
    number++;
  }
  //
  if (ounit!=NULL) *ounit=unit;
  if (oform!=NULL) *oform=form;
  if (odir!=NULL) *odir=ExtractFilePath(fn);
}



class TComponentLooker : public TObject
{ public:
  TFormScanner *scanner;
  __fastcall TComponentLooker() : TObject() {scanner=new TFormScanner(); scanner->OnScanModule=ScanModule; scanner->OnScanForm=ScanForm; scanner->OnScanComponent=ScanComponent;}
  __fastcall ~TComponentLooker() { if (scanner!=NULL) delete scanner; scanner=NULL;}
  //
  AnsiString searchforform, searchforunit, searchforcname, searchforctype;
  AnsiString foundform,foundunit,foundcname,foundctype;
  AnsiString currentform, currentunit, currentcname, currentctype;
  bool foundit;
  bool __fastcall Look(_di_IOTAProject p,AnsiString *aname,AnsiString *atype,AnsiString *aform,AnsiString *aunit)
  { searchforform=""; if (aform!=NULL) searchforform=*aform;
    searchforunit=""; if (aunit!=NULL) searchforunit=*aunit;
    searchforcname=""; if (aname!=NULL) searchforcname=*aname;
    searchforctype=""; if (atype!=NULL) searchforctype=*atype;
    *aform=""; *aunit=""; *aname=""; *atype="";
    foundform=""; foundunit=""; foundcname=""; foundctype="";
    foundit=false;
    // first scan: search the given project (or the current one if it's empty)
    if (p!=NULL) scanner->autoproject=p; else scanner->autoproject=GetCurrentProject();
    scanner->Scan();
    scanner->autoproject=NULL;
    if (foundit)
    { *aname=foundcname; *atype=foundctype; *aform=foundform; *aunit=foundunit;
      return true;
    }
    // second scan: maybe scan all the projects
    _di_IOTAModuleServices ms=GetModuleServices(); int numprojects=0;
    for (int i=0; i<ms->ModuleCount; i++)
    { _di_IOTAProject p=GetModuleAsProject(ms->Modules[i]);
      if (p!=NULL) numprojects++;
    }
    if (numprojects==1) return false;
    scanner->Scan();
    if (foundit)
    { *aname=foundcname; *atype=foundctype; *aform=foundform; *aunit=foundunit;
      return true;
    }
    return false;
  }
  void __fastcall ScanModule(AnsiString fn,AnsiString un,AnsiString fo,AnsiString dc,bool &cancel)
  { if (foundit) {cancel=true;return;}
    if (searchforunit!="" && searchforunit!=un) cancel=true;
    currentunit=fn;
  }
  void __fastcall ScanForm(TCustomForm *form,AnsiString &s,bool &cancel)
  { if (foundit) {cancel=true; return;}
    if (searchforform!="" && searchforform!=s) {cancel=true;return;}
    currentform=s;
  }
  void __fastcall ScanComponent(TCustomForm *form,TComponent *com,AnsiString &s,bool &cancel)
  { if (foundit) {cancel=true; return;}
    if (searchforcname!="" && searchforcname!=com->Name) {cancel=true; return;}
    if (searchforctype!="" && searchforctype!=com->ClassName()) {cancel=true; return;}
    currentcname=com->Name;
    currentctype=com->ClassName();
    foundform=currentform; foundunit=currentunit; foundcname=currentcname; foundctype=currentctype;
    foundit=true;
    cancel=true;
  }
};
bool __fastcall LookForComponent(_di_IOTAProject p,AnsiString *name,AnsiString *type, AnsiString *form,AnsiString *unit)
{ TComponentLooker *cl=new TComponentLooker(); bool res=cl->Look(p,name,type,form,unit); delete cl; return res;
}













//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// PROPERTY MANIPULATION ROUTINES
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// Ways to manipulate properties as strings

class TComponentPropertyTextData : public TObject
{ public: AnsiString prop; AnsiString val; bool ok; bool set;
  TComponentPropertyTextData(AnsiString p,AnsiString v,bool s) {prop=p; val=v; set=s; ok=false;}
  void __fastcall Proc(TPropertyEditor *pe)
  { if (ok) return;
    AnsiString sp, rp; int pos=prop.Pos(".");
    if (pos==0) {sp=prop; rp="";}
    else {sp=prop.SubString(1,pos-1); rp=prop.SubString(pos+1,prop.Length()-pos);}
    AnsiString pn=pe->GetName(); if (pn!=sp) return;
    if (pos==0) {if (set) pe->SetValue(val); else val=pe->GetValue(); ok=true; return;}
    else {prop=rp; pe->GetProperties(Proc);}
  }
};

AnsiString __fastcall ComponentPropertyTextOp(_di_IFormDesigner ides,_di_IPersistent ip,AnsiString prop,AnsiString val,bool set)
{ ip->AddRef(); TPersistent *p=TryExtractPersistent(ip); // manual reference counting
  TDesignerSelectionList *list=new TDesignerSelectionList(); list->Add(p);
  TTypeKinds kinds; kinds << tkUnknown << tkInteger << tkChar << tkEnumeration << tkFloat
    << tkString << tkSet << tkClass << tkMethod << tkWChar << tkLString << tkWString
    << tkVariant << tkArray << tkRecord << tkInterface;
  TComponentPropertyTextData *dat=new TComponentPropertyTextData(prop,val,set);
  GetComponentProperties(list,kinds,ides,dat->Proc);
  AnsiString res=dat->val;
  delete dat;
  delete list;
  ip->Release();
  return res;
}
AnsiString __fastcall ComponentPropertyTextOpOTA(_di_IOTAFormEditor fe,_di_IOTAComponent c,AnsiString prop,AnsiString val,bool set)
{ _di_INTAFormEditor nfe=NULL;
  HRESULT res=fe->QueryInterface(__uuidof(INTAFormEditor),(void**)&nfe);
  if (FAILED(res)) return "";
  _di_IFormDesigner ides=nfe->GetFormDesigner();
  _di_IPersistent ip=c->GetIPersistent();
  return ComponentPropertyTextOp(ides,ip,prop,val,set);
}
AnsiString __fastcall GetComponentPropertyTextOTA(_di_IOTAFormEditor fe,_di_IOTAComponent c,AnsiString prop)
{ return ComponentPropertyTextOpOTA(fe,c,prop,"",false);
}
AnsiString __fastcall GetComponentPropertyText(_di_IFormDesigner ides,_di_IPersistent ip,AnsiString prop)
{ return ComponentPropertyTextOp(ides,ip,prop,"",false);
}
void __fastcall SetComponentPropertyTextOTA(_di_IOTAFormEditor fe,_di_IOTAComponent c,AnsiString prop,AnsiString val)
{ ComponentPropertyTextOpOTA(fe,c,prop,val,true);
}
void __fastcall SetComponentPropertyText(_di_IFormDesigner ides,_di_IPersistent ip,AnsiString prop,AnsiString val)
{ ComponentPropertyTextOp(ides,ip,prop,val,true);
}
AnsiString __fastcall GetComponentNameOTA(_di_IOTAFormEditor fe,_di_IOTAComponent c)
{ return GetComponentPropertyTextOTA(fe,c,"Name");
}
AnsiString __fastcall GetComponentName(_di_IFormDesigner ides,_di_IPersistent ip)
{ return GetComponentPropertyText(ides,ip,"Name");
}

bool __fastcall ExecuteComponentVerb(_di_IFormDesigner ides,_di_IPersistent ip,AnsiString verb)
{ TComponent *com=ExtractComponent(ip); if (com==NULL) {db("ExecuteComponentVerb. Couldn't get TComponent from IPersistent.");return false;}
  TComponentEditor *ce=GetComponentEditor(com,ides);
  if (ce==NULL) {db("ExtractComponentVerb. Couldn't find the component-editor for component "+com->Name+dptr(com));return false;}
  for (int i=0; i<ce->GetVerbCount(); i++)
  { if (verb==ce->GetVerb(i))
    { db("ExecuteComponentVerb - executing verb "+verb+" on component "+com->Name+dptr(com));
      ce->ExecuteVerb(i);
      db("ExecuteComponentVerb - finished verb");
      return true;
    }
  }
  dbe("ExecuteComponentVerb - couldn't find verb "+verb+" in component "+com->Name+dptr(com));
  return false;
}
bool __fastcall ExecuteComponentVerbOTA(_di_IOTAFormEditor fe,_di_IOTAComponent c,AnsiString verb)
{ _di_IDesigner des=GetDesignerFromFormEditor(fe); if (des==NULL) {db("ExecuteComponentVerb. Couldn't get IDesigner from IOTAFormEditor");return false;}
  _di_IPersistent ip=c->GetIPersistent(); if (ip==NULL) {db("ExecuteComponentVerb. Couldn't get IPersistent from IOTAComponent");return false;}
  return ExecuteComponentVerb(des,ip,verb);
}



void __fastcall SetComponentPropertyImage(_di_IFormDesigner ides,_di_IPersistent ip,AnsiString fn,bool isbmp)
{ if (fn.Length()==0) return;
  Graphics::TBitmap *srcb=new Graphics::TBitmap; TIcon *srci=new TIcon;
  bool okay=true;
  try
  { if (isbmp)
    { if (fn[1]=='*') srcb->LoadFromResourceName((int)HInstance,fn.SubString(2,fn.Length()-1));
      else srcb->LoadFromFile(fn);
    }
    else
    { if (fn[1]=='*')
      { HICON hIcon=LoadIcon(HInstance,fn.SubString(2,fn.Length()-1).c_str());
        if (hIcon==NULL) throw new Exception("Not a valid resource icon");
        HICON hold=srci->ReleaseHandle(); srci->Handle=hIcon; DestroyIcon(hold);
      }
      else srci->LoadFromFile(fn);
    }
  } catch (...) {okay=false;}
  //
  if (okay)
  { if (ip->IsComponent())
    { ip->AddRef(); TComponent *c=TryExtractComponent(ip);
      TImage *image=dynamic_cast<TImage*>(c);
      if (image!=NULL)
      { if (isbmp) image->Picture->Bitmap->Assign(srcb);
        else image->Picture->Icon->Assign(srci);
      }
    }
  }
  delete srcb; delete srci;
  ip->Release();
  return;
}
void __fastcall SetComponentImageOTA(_di_IOTAFormEditor fe,_di_IOTAComponent c,AnsiString fn,bool isbmp)
{ _di_INTAFormEditor nfe=NULL;
  HRESULT res=fe->QueryInterface(__uuidof(INTAFormEditor),(void**)&nfe);
  if (FAILED(res)) return;
  _di_IFormDesigner ides=nfe->GetFormDesigner();
  _di_IPersistent ip=c->GetIPersistent();
  return SetComponentPropertyImage(ides,ip,fn,isbmp);
}







//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// RESOURCE FUNCTIONS
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// The following code is from Lischner ch. 16. He says:
// Note that the data format in a file differs from the data
// format in the resource. The ResInfo unit declares the format
// for resource files. The declarations below pertain to
// external .ICO and .CUR files.
#pragma pack(push,1)
typedef struct tagFileGroupItem
{ BYTE bWidth;
  BYTE bHeight;
  BYTE bColorCount;
  BYTE bReserved;
  WORD wPlanes; WORD wBitCount;
  DWORD dwSize; // item size in bytes
  DWORD dwOffset; // start of item in the file
} TFileGroupItem;

typedef struct tagFileGroupHeader
{ WORD wReserved;
  WORD wType; // 1 for icons, 2 for cursors
  WORD wCount;
  TFileGroupItem Items[1];
} TFileGroupHeader;

typedef struct tagGroupItem
{ BYTE bWidth;
  BYTE bHeight;
  BYTE bColorCount;
  BYTE bReserved;
  WORD wPlanes;
  WORD wBitCount;
  DWORD dwBytesInRes;
  WORD wNameOrdinal;
} TGroupItem;

typedef struct tagGroupHeader
{ WORD wReserved;
  WORD wType; // 1 for icons, 2 for cursors
  WORD wCount;
  TGroupItem Items[1];
} TGroupHeader;
#pragma pack(pop)

// RESOURCE STREAMS.
// Just wrappers around a resource entry,
// and around a file (use "*name" if the file has been included as RT_RCDATA,
// or just "name.ext" if it's on disk)
class TResourceDataStream : public TCustomMemoryStream
{
public:
  __fastcall TResourceDataStream(_di_IOTAResourceEntry re) : TCustomMemoryStream()
  { SetPointer(re->GetData(),re->DataSize);
  }
  int __fastcall Write(const void*Buffer,int Count)
  { if (Position+Count>Size) Count=Size-Position;
    Move(Buffer,((char*)Memory)+Position,Count);
    Position=Position+Count;
    return Count;
  }
};
class TFileDataStream : public TMemoryStream
{
public:
  __fastcall TFileDataStream(AnsiString fn) : TMemoryStream()
  { if (fn[1]=='*')
    { TResourceStream *InStream=new TResourceStream((int)HInstance,fn.SubString(2,fn.Length()-1),RT_RCDATA);
      CopyFrom(InStream,0); delete InStream;
    }
    else
    { TFileStream *InStream=new TFileStream(fn,fmOpenRead|fmShareDenyWrite);
      CopyFrom(InStream, 0);
      delete InStream;
    }
  }
};



bool __fastcall DeleteIconGroup(char *name,_di_IOTAProjectResource pr)
{ if (name==NULL) return false;
  if (pr==NULL) pr=GetProjectResources(NULL);
  if (pr==NULL) return false;
  char *type=RT_GROUP_ICON;
  _di_IOTAResourceEntry re=pr->FindEntry(type,name);
  if (re==NULL) return false;
  void*grouph=re->GetEntryHandle();
  TResourceDataStream *ds=new TResourceDataStream(re); int size=re->GetDataSize();
  char *c=new char[size]; ds->ReadBuffer(c,size); delete ds;
  //
  TGroupHeader *ResHdr=(TGroupHeader*)c;
  for (int i=0; i<ResHdr->wCount; i++)
  { int IconNum = ResHdr->Items[i].wNameOrdinal;
    char *iconname = MAKEINTRESOURCE(IconNum);
    char *icontype = RT_ICON;
    _di_IOTAResourceEntry icone = pr->FindEntry(icontype,iconname);
    if (icone==NULL) Application->MessageBox((AnsiString("Suspiciously didn't find icon ")+AnsiString(IconNum)+" of the group").c_str(),"Unexpected error",MB_OK);
    else pr->DeleteEntry(icone->GetEntryHandle());
  }
  delete[] c;
  pr->DeleteEntry(grouph);
  return true;
}


bool __fastcall FileOrResourceExists(AnsiString fn)
{ if (fn.Length()==0) return false;
  if (fn[1]!='*') return FileExists(fn);
  HRSRC hrsrc=FindResource(HInstance,fn.SubString(2,fn.Length()-1).c_str(),RT_RCDATA);
  if (hrsrc==NULL) return false;
  HGLOBAL hglob=LockResource(hrsrc);
  if (hglob==NULL) return false;
  return true;
}


_di_IOTAResourceEntry __fastcall AddIconFileResource(AnsiString fn, char *name,WORD flags,WORD langid,int DataVersion,int Version,int Characteristics,_di_IOTAProjectResource pr)
{ if (!FileOrResourceExists(fn)) {dbe("AddIconFileResource. The file "+fn+" doesn't exist.");return NULL;}
  if (pr==NULL) pr=GetProjectResources(NULL);
  if (pr==NULL) {dbe("AddIconFileResource. Failed to get project resources.");return NULL;}
  // Lischner: Create an icon or cursor resource by separating the group header
  // and creating a separate resource for each item in the group. Then create a
  // group entry. Automatically assign a resource identifier for each individual
  // resource, ensuring that the identifiers do not conflict with any identifiers already in use.
  // First determine the highest resource ID currently assigned to a resource of this type
  int MaxId=0;
  for (int i=0; i<pr->GetEntryCount(); i++)
  { _di_IOTAResourceEntry re=pr->GetEntry(i);
    if (re->GetResourceType()==RT_ICON)
    { int resid=(int)re->GetResourceName();
      if (HIWORD(resid)==0 && resid>MaxId) MaxId=resid;
    }
  }
  // Assign MaxId+1 as the ID of the first image resource.
  MaxId++;
  TFileDataStream *fs=new TFileDataStream(fn);
  // Extract each icon/cursor image.
  TFileGroupHeader *FileHdr = (TFileGroupHeader*)fs->Memory;
  int ResHdrSize = 3*sizeof(WORD) + FileHdr->wCount*sizeof(TGroupItem);
  TGroupHeader *ResHdr=(TGroupHeader*)new char[ResHdrSize];
  ResHdr->wReserved=0;
  ResHdr->wType=FileHdr->wType;
  ResHdr->wCount=FileHdr->wCount;
  for (int i=0; i<FileHdr->wCount; i++)
  { CopyMemory(&(ResHdr->Items[i]),&(FileHdr->Items[i]),sizeof(TGroupItem));
    ResHdr->Items[i].wNameOrdinal=(WORD)MaxId;
    _di_IOTAResourceEntry re=pr->CreateEntry(RT_ICON,MAKEINTRESOURCE(MaxId),flags,langid,DataVersion,Version,Characteristics);
    re->SetDataSize(ResHdr->Items[i].dwBytesInRes);
    TResourceDataStream *ds=new TResourceDataStream(re);
    fs->Position = FileHdr->Items[i].dwOffset;
    ds->CopyFrom(fs, FileHdr->Items[i].dwSize);
    delete ds;
    MaxId++;
  }
  // Create the group resource entry.
  _di_IOTAResourceEntry re=pr->CreateEntry(MAKEINTRESOURCE(RT_GROUP_ICON),name,flags,langid,DataVersion,Version,Characteristics);
  re->SetDataSize(ResHdrSize);
  TResourceDataStream *ds=new TResourceDataStream(re);
  ds->Write(ResHdr,ResHdrSize); delete ds;
  delete[] ResHdr;
  delete fs;
  return re;
}

_di_IOTAResourceEntry __fastcall AddBitmapFileResource(AnsiString fn,char *name,WORD flags,WORD langid,int DataVersion,int Version,int Characteristics,_di_IOTAProjectResource pr)
{ if (!FileOrResourceExists(fn)) {dbe("AddBitmapFileResource. The file "+fn+" doesn't exist.");return NULL;}
  if (pr==NULL) pr=GetProjectResources(NULL);
  if (pr==NULL) {dbe("AddBitmapFileResource. Failed to get project resources.");return NULL;}
  // Lischner: When storing a bitmap in a resource file we do not store the
  // bitmap file header that exists in the disk file.
  _di_IOTAResourceEntry re=pr->CreateEntry(RT_BITMAP,name,flags,langid,DataVersion,Version,Characteristics);
  TFileDataStream *fs=new TFileDataStream(fn);
  re->SetDataSize(fs->Size-sizeof(TBitmapFileHeader));
  fs->Seek(sizeof(TBitmapFileHeader),soFromBeginning);
  TResourceDataStream *ds=new TResourceDataStream(re);
  ds->CopyFrom(fs,fs->Size-fs->Position);
  delete ds; delete fs;
  return re;
}


















//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// FORMSCANNER
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// TFormScanner for scanning all forms and components in the current package.
__fastcall TFormScanner::TFormScanner() : TObject(), OnScanProject(NULL), OnScanHiddenForm(NULL), OnScanModule(NULL), OnScanForm(NULL), OnScanComponent(NULL), OnScanSubComponent(NULL), results(NULL), res2(NULL), autoproject(NULL)
{ autoforms=false; autocomponents=false; autohiddens=false; dosubcomponents=false;
  GetStrProc=NULL;
}
__fastcall TFormScanner::~TFormScanner()
{ autoproject=NULL;
  if (results!=NULL) delete results; results=NULL;
  if (res2!=NULL) delete res2; res2=NULL;
}
bool __fastcall TFormScanner::StringStandsForHiddenForm(AnsiString s)
{ int len=s.Length();
  if (len>=5 && s[1]=='(' && s.SubString(len-3,4)=="...)") return true;
  return false;
}
void __fastcall TFormScanner::Add(AnsiString n,TObject *o,AnsiString s)
{ results->AddObject(n,o); res2->Add(s);
}

void __fastcall TFormScanner::QuickScan(_di_IFormDesigner ides,TGetStrProc proc,bool af,bool ac,bool ah)
{ autoproject = GetProjectFromDesigner(ides);
  GetStrProc=proc;
  autoforms=af;
  autocomponents=ac;
  autohiddens=ah;
  Scan();
}

void __fastcall TFormScanner::Scan()
{ if (results!=NULL) delete results; if (res2!=NULL) delete res2;
  results=new TStringList(); res2=new TStringList();
  _di_IOTAModuleServices ms=GetModuleServices();
  for (int i=0; i<ms->ModuleCount; i++)
  { _di_IOTAProject p=GetModuleAsProject(ms->Modules[i]);
    if (p!=NULL) ScanProject(p);
  }
}

void __fastcall TFormScanner::ScanProject(_di_IOTAProject p)
{ if (autoproject!=NULL && autoproject->GetFileName()!=p->GetFileName()) return;
  bool cancel=false;
  if (OnScanProject!=NULL) OnScanProject(p->GetFileName(),cancel);
  if (cancel) return;
  for (int i=0; p!=NULL && i<p->GetModuleCount(); i++)
  { ScanModule(p->GetModule(i));
  }
}

void __fastcall TFormScanner::ScanModule(_di_IOTAModuleInfo mi)
{ bool cancel=false;
  if (OnScanModule!=NULL) OnScanModule(mi->GetFileName(),mi->GetName(),mi->GetFormName(),mi->GetDesignClass(),cancel);
  if (cancel) return;
  // now to its form, if it has one
  if (mi->GetModuleType()!=0) return;
  _di_IOTAModule m=GetModuleServices()->FindModule(mi->GetFileName());
  if (m==NULL) // Maybe the module isn't yet live
  { AnsiString s="("+mi->GetFormName()+"...)";
    bool cancel=false;
    if (OnScanHiddenForm!=NULL) OnScanHiddenForm(mi->GetFormName(),mi->GetName(),s,cancel);
    if (cancel) return;
    if (autohiddens)
    { Add(s,NULL,mi->GetFileName());
      if (GetStrProc!=NULL) GetStrProc(s);
    }
    return;
  }
  _di_IOTAFormEditor fe=GetFormEditorForModule(m); if (fe==NULL) {dbe("FormScanner. Couldn't get form editor when scanning a live module.");return;}
  ScanForm(fe);
}


void __fastcall TFormScanner::ScanForm(_di_IOTAFormEditor fe)
{ _di_IOTAComponent ic = fe->GetRootComponent(); if (ic==NULL) {db("FormScanner. Failed to get root component of fe.");return;}
  TComponent *c=(TComponent*)ic->GetComponentHandle(); if (c==NULL) {db("FormScanner. Failed to get root TComponent of fe.");return;}
  TCustomForm *form=dynamic_cast<TCustomForm*>(c); if (form==NULL) {db("FormScanner. Failed to get CustomForm of fe.");return;}
  AnsiString s=form->Name;
  bool cancel=false;
  if (OnScanForm!=NULL) OnScanForm(form,s,cancel);
  if (cancel) return;
  // maybe list it
  if (autoforms)
  { Add(s,form,"");
    if (GetStrProc!=NULL) GetStrProc(s);
  }
  // enumerate immediate children
  ScanComponent(fe,ic,0,"");
}

void __fastcall TFormScanner::ScanComponent(_di_IOTAFormEditor fe,_di_IOTAComponent c,int depth,AnsiString path)
{ TCustomForm *form=(TCustomForm*)fe->GetRootComponent()->GetComponentHandle(); if (form==NULL) {db("FormScanner. Scanning component got a NULL form.");return;}
  TComponent *com=(TComponent*)c->GetComponentHandle(); if (com==NULL) {db("FormScanner. Got a NULL child component.");return;}
  AnsiString s=path+"->"+com->Name;
  bool cancel=false;
  if (depth==1 && OnScanComponent!=NULL) OnScanComponent(form,com,s,cancel);
  if (depth>1 && OnScanComponent!=NULL) OnScanSubComponent(form,com,s,cancel);
  if (cancel) return;
  // maybe list it
  if (autocomponents && depth==1)
  { Add(s,com,"");
    if (GetStrProc!=NULL) GetStrProc(s);
  }
  // maybe subcomponents?
  if (depth>=1 && !dosubcomponents) return;
  for (int i=0; i<c->GetComponentCount(); i++)
  { _di_IOTAComponent d=c->GetComponent(i); if (d==NULL) {db("FormScanner. Got a NULL child component.");return;}
    AnsiString s=com->Name; if (depth>0) s=path+"->"+s;
    ScanComponent(fe,d,depth+1,s);
  }
}



bool __fastcall TFormScanner::OpenFormIfHidden(AnsiString val)
{ if (!StringStandsForHiddenForm(val)) return false;
  AnsiString un="";
  int i=results->IndexOf(val);
  if (i!=-1)
  { AnsiString fo= val.SubString(2, val.Length()-5);
    AnsiString un=res2->Strings[i];
    _di_IOTAFormEditor fe=GetFormEditorEx(fo,un); // forces the thing to open!
    if (fe!=NULL) fe->Show();
  }
  return true;
}














//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// PROPERTY AND COMPONENT EDITORS
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// TSwappingComponentEditor -- has the verb 'Swap...' to swap the component for another
// TPropertyWatcher -- base class for something that might auto-update a property value
// TFarPropertyWatcher -- auto-updating based on NameNotifications
// GetPropertyWatcher -- given a property-editor on particular component, returns its property-watcher
// TFormScanningEditor -- base class for a property editor that looks at all forms
// TComponentNameEditor -- prop.ed for text-properties with values 'form1->button1'
// TFormNameEditor -- prop.ed for text-properties with values 'form1'
// TFormTypeEditor -- prop.ed for text-properties with values 'Tform1'

__fastcall TSwappingComponentEditor::TSwappingComponentEditor(TComponent* Component, _di_IFormDesigner Designer) : TComponentEditor(Component,Designer), hmenu(NULL), names(NULL), types(NULL)
{
}
__fastcall TSwappingComponentEditor::~TSwappingComponentEditor()
{
}
int __fastcall TSwappingComponentEditor::GetVerbCount()
{ return 1+TComponentEditor::GetVerbCount();
}
AnsiString __fastcall TSwappingComponentEditor::GetVerb(int i)
{ if (i==0) return GetSwapText(); else return TComponentEditor::GetVerb(i-1);
}
void __fastcall TSwappingComponentEditor::Edit()
{ dbe("Unexpectedly called the SwappingComponent editor");
}
void __fastcall TSwappingComponentEditor::ExecuteVerb(int i)
{ if (i!=0) TComponentEditor::ExecuteVerb(i-1);
  try
  { names=new TStringList(); types=new TStringList(); hmenu=CreatePopupMenu();
    ExecuteSwapMenu();
    DestroyMenu(hmenu); hmenu=NULL; delete names; names=NULL; delete types; types=NULL;
  } catch(Exception &e) {dbee("Error executing SwappingComponent swap menu",e);}
}
int __fastcall TSwappingComponentEditor::AddMenuItem(AnsiString s,AnsiString n)
{ int id=names->Count+100;
  MENUITEMINFO mi; ZeroMemory(&mi,sizeof(mi)); mi.cbSize=sizeof(mi);
  mi.fMask=MIIM_DATA|MIIM_ID|MIIM_STATE|MIIM_TYPE; mi.fState=MFS_ENABLED;
  if (s=="-") {mi.fType=MFT_SEPARATOR; id=0;} else mi.fType=MFT_STRING;
  mi.wID=id; mi.dwTypeData=s.c_str(); mi.cch=s.Length();
  InsertMenuItem(hmenu,names->Count,TRUE,&mi);
  types->Add(s); names->Add(n);
  return id-100;
}
int __fastcall TSwappingComponentEditor::ExecuteMenu()
{ _di_IComponent ic=GetIComponent();
  POINT pt; pt.x=ic->DesignInfo.x; pt.y=ic->DesignInfo.y;
  pt=Designer->ContainerWindow->ClientToScreen(pt);
  int res= TrackPopupMenu(hmenu,TPM_LEFTALIGN|TPM_TOPALIGN|TPM_RETURNCMD,pt.x,pt.y,0,Designer->ContainerWindow->Handle,NULL);
  return res-100;
}
_di_IOTAComponent __fastcall TSwappingComponentEditor::ExecuteSwap(int i)
{ if (i<0 || i>=names->Count) return NULL;
  AnsiString cname=names->Strings[i]; AnsiString ctype=types->Strings[i];
  if (ctype=="-") return NULL;
  // The suggested cname will be "" pick one, or "Sugg1" use sugg name, or "Sugg" without an index
  int number=0; bool isalreadytaken=true;
  AnsiString cbase=cname; bool isnumbered=ExtractLastNumber(&cname,&number);
  while (isalreadytaken)
  { isalreadytaken=false;
    cname=cbase; if (isnumbered) cname=cname+AnsiString(number);
    isalreadytaken = (Designer->GetComponent(cname) != NULL);
    number++;
  }
  // We have to go to IOTA to create the thing. That's because we need to
  // create a component given an AnsiString of its type.
  _di_IOTAModule m = GetModuleContainingComponent(Component); if (m==NULL) {dbe("swap. Unable to find module containing new component");return NULL;}
  _di_IOTAFormEditor fe = GetFormEditorForModule(m); if (fe==NULL) {dbe("swap. Unable to find form-editor containing new component");return NULL;}
  _di_IOTAComponent prev = fe->GetComponentFromHandle(Component); if (prev==NULL) {dbe("swap. Unable to find component on form");return NULL;}
  _di_IOTAComponent par = prev->GetParent(); if (par==NULL) par=fe->GetRootComponent();
  int x=GetIComponent()->DesignInfo.x, y=GetIComponent()->DesignInfo.y, w=32, h=32;
  _di_IOTAComponent c = CreateComponent(fe,par,ctype,cname,x,y,w,h);
  if (c!=NULL) prev->Delete(); // remove the old one
  return c;
}
_di_IOTAComponent __fastcall TSwappingComponentEditor::CreateComponent(_di_IOTAFormEditor fe,_di_IOTAComponent par,AnsiString type,AnsiString name,int x,int y,int w,int h)
{ _di_IOTAComponent c = fe->CreateComponent(par,type,x,y,w,h); if (c==NULL) {dbe("swap. Failed to create '"+type+"' component");return NULL;}
  SetComponentPropertyTextOTA(fe,c,"Name",name);
  c->Focus(false);
  return c;
}

AnsiString __fastcall TSwappingComponentEditor::GetSwapText()
{ return "Swap...";
}















TPropertyWatcher* __fastcall GetPropertyWatcher(TPropertyEditor *pe,int i)
{ // Recall that the PropertyWatcher was created as a child of the component...
  TComponent *com=dynamic_cast<TComponent*>(pe->GetComponent(i));
  if (com!=NULL)
  { for (int i=0; i<com->ComponentCount; i++)
    { TPropertyWatcher *pw=dynamic_cast<TPropertyWatcher*>(com->Components[i]);
      if (pw!=NULL && pw->propname==pe->GetName()) return pw;
    }
  }
  return NULL;
}


__fastcall TPropertyWatcher::TPropertyWatcher(TComponent *Owner) : TComponent(Owner) {}
void __fastcall TPropertyWatcher::OnExists(TPropertyEditor *pe) {propname=pe->GetName();}
void __fastcall TPropertyWatcher::OnSet(TPropertyEditor *pe,AnsiString sold,AnsiString snew) {}


__fastcall TFarPropertyWatcher::TFarPropertyWatcher(TComponent *Owner) : TPropertyWatcher(Owner), ignorethis(false), prefix("")
{ AnsiString s="fpw"; if (Owner!=NULL) s=s+Owner->Name;
  SafeSetName(Owner,this,s);
}
void __fastcall TFarPropertyWatcher::Watch(AnsiString s)
{ db("fpw. "+Name+dptr(this)+"watch = "+s);
  NameNotification(s,OnName);
  NameNotification("<project>0x"+IntToHex((int)Owner,8),OnName); // in the same project as our client
}
AnsiString __fastcall TFarPropertyWatcher::Decorate(AnsiString s) {if (s=="") return s; else return prefix+s;}
AnsiString __fastcall TFarPropertyWatcher::Undecorate(AnsiString s) {if (s=="") return s; else return s.SubString(prefix.Length()+1,s.Length()-prefix.Length());}
void __fastcall TFarPropertyWatcher::OnExists(TPropertyEditor *pe)
{ SafeSetName(Owner,this,Name+pe->GetName());
  db("fpw. "+Name+dptr(this)+": OnExists");
  TPropertyWatcher::OnExists(pe); Watch(Undecorate(pe->GetValue()));
}
void __fastcall TFarPropertyWatcher::OnSet(TPropertyEditor *pe,AnsiString sold,AnsiString snew)
{ if (ignorethis) return;
  db("fpw. "+Name+dptr(this)+"OnSet "+sold+" -> "+snew);
  Watch(Undecorate(snew));
}
void __fastcall TFarPropertyWatcher::OnName(_di_IOTAProject notp,_di_IOTAModule notm,_di_IOTAFormEditor cfe,_di_IOTAComponent cc,AnsiString name,AnsiString oldname,TWatchAction wa,TWatchSubject ws)
{ // Here's a thorny problem. If this property points to the form that the
  // component is on, and we are being told that the form is deleted, then
  // it's too late for us to get a form-editor! or if we're being told that
  // we ourselves are being deleted, then that is also too late.
  if (Owner==NULL) return; // we seemed to be getting notifications with NULL owner!
  TComponent *own=Owner; while (own->Owner!=NULL) own=own->Owner;
  AnsiString scom=Owner->Name; AnsiString sform=own->Name; if (Owner->Owner==NULL) scom="";
  if (name=="" && oldname==sform) return; // we point to our own form which is being deleted.
  if (name=="" && oldname==sform+"."+scom) return; // we point to ourself and are being deleted.
  ignorethis=true;
  db("fpw. "+Name+dptr(this)+"OnName "+oldname+" -> "+name);
  _di_IOTAModule m=GetModuleContainingComponent(Owner); if (m==NULL) {dbe("fpw. Failed to find owner's module "+Owner->Name+"->"+propname);return;}
  _di_IOTAFormEditor fe = GetFormEditorForModule(m); if (fe==NULL) {dbe("fpw. Failed to find owner's formeditor "+Owner->Name+"->"+propname);return;}
  _di_IDesigner id = GetDesignerFromFormEditor(fe); if (id==NULL) {dbe("fpw. Failed to find owner's idesigner "+Owner->Name+"->"+propname);return;}
  _di_IPersistent ip = MakeIPersistent(Owner);  if (id==NULL) {dbe("Failed to find owner's ipersistent "+Owner->Name+"->"+propname);return;}
  SetComponentPropertyText(id,ip,propname,Decorate(name));
  ignorethis=false;
}








__fastcall TFormScanningEditor::TFormScanningEditor(const _di_IFormDesigner ides,int count) : TPropertyEditor(ides,count), prefix("")
{ scanner=new TFormScanner();
}
__fastcall TFormScanningEditor::~TFormScanningEditor()
{ if (scanner!=NULL) delete scanner; scanner=NULL;
}

void __fastcall TFormScanningEditor::GetValues(TGetStrProc Proc)
{ GenerateList();
  for (int i=0; i<scanner->results->Count; i++)
  { AnsiString s=scanner->results->Strings[i];
    Proc(s);
  }
}
TPropertyAttributes __fastcall TFormScanningEditor::GetAttributes()
{ TPropertyAttributes pa=TPropertyEditor::GetAttributes();
  pa << paValueList << paSortList;
  return pa;
}

AnsiString __fastcall TFormScanningEditor::GetValue()
{ return RetrieveValue();
}
void __fastcall TFormScanningEditor::SetValue(AnsiString NewValue)
{ GenerateList();
  if (scanner->OpenFormIfHidden(NewValue)) return;
  AnsiString OldValue=GetValue();
  StoreValue(NewValue); Modified();
  for (int i=0; i<PropCount; i++)
  { TPropertyWatcher *pw=GetPropertyWatcher(this,i);
    if (pw!=NULL) pw->OnSet(this,OldValue,NewValue);
  }
}





__fastcall TComponentNameEditor::TComponentNameEditor(const _di_IFormDesigner ides,int count) : TFormScanningEditor(ides,count) {}
void __fastcall TComponentNameEditor::GenerateList()
{ scanner->QuickScan(Designer,NULL,false,true,true);
};
void __fastcall TComponentNameEditor::OnExists(TPropertyEditor *pe)
{ TComponent *c=dynamic_cast<TComponent*>(pe->GetComponent(0)); if (c==NULL) return;
  TFarPropertyWatcher *fpw=new TFarPropertyWatcher(c); fpw->OnExists(pe);
}





__fastcall TFormNameEditor::TFormNameEditor(const _di_IFormDesigner ides,int count) : TFormScanningEditor(ides,count)
{ scanner->OnScanModule=this->OnScanModule;
}
void __fastcall TFormNameEditor::GenerateList()
{ scanner->QuickScan(Designer,NULL,false,false,false);
}
void __fastcall TFormNameEditor::OnScanModule(AnsiString fn,AnsiString un,AnsiString fo,AnsiString dc,bool &cancel)
{ if (fo!="") scanner->Add(fo,NULL,"");
}
void __fastcall TFormNameEditor::OnExists(TPropertyEditor *pe)
{ TComponent *c=dynamic_cast<TComponent*>(pe->GetComponent(0)); if (c==NULL) return;
  TFarPropertyWatcher *fpw=new TFarPropertyWatcher(c); fpw->OnExists(pe);
}






__fastcall TFormTypeEditor::TFormTypeEditor(const _di_IFormDesigner ides,int count) : TFormScanningEditor(ides,count)
{ scanner->OnScanModule=this->OnScanModule;
}
void __fastcall TFormTypeEditor::GenerateList()
{ scanner->QuickScan(Designer,NULL,false,false,false);
}
void __fastcall TFormTypeEditor::OnScanModule(AnsiString fn,AnsiString un,AnsiString fo,AnsiString dc,bool &cancel)
{ if (fo!=NULL) scanner->Add("T"+fo,NULL,"");
}
void __fastcall TFormTypeEditor::OnExists(TPropertyEditor *pe)
{ TComponent *c=dynamic_cast<TComponent*>(pe->GetComponent(0)); if (c==NULL) return;
  TFarPropertyWatcher *fpw=new TFarPropertyWatcher(c); fpw->prefix="T"; fpw->OnExists(pe);
}


















//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// THEWATCHER
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// This provides the watching-services
// When the design-time package loads it creates TWatchExpert, whose sole purpose
// is to create a TTheWatcher. (and destroy it when the expert is unloaded).
// (1) Any client component can ask TheWatcher for notifications. TheWatcher
// keeps a list of all such clients. (if you delete a client, TheWatcher will
// remove it from its list automatically). You might create your client owned
// by TTheWatcher (for automatic cleanup); or you might create it owned by
// some other component (so the watcher gets deleted automatically when that
// other component is deleted).
// (2) TheWatcher creates an IDENotifier, which in turn creates ModuleNotifiers
// for all open modules, which in turn create FormNotifiers if appropriate. With
// the FormNotifiers it gets told if any component is renamed.
// (3) There are a few special kinds of watches already written, which you
// can use with RegisterPlonkVerb and RegisterPropertyWatcher and
// NameNotification. Or you can write your own watches.
//
// IMPLEMENTATION NOTES
//
// (1) Reference counting is hard! Imagine we had done AddNotifier() on
// something. In a normal C++ program you'd make the thing's destructor
// do the ReleaseNotifier(). That style of programming doesn't work any more.
// That's because the destructor never gets called until all references
// to it have been released; and the AddNotifier itself causes a reference!
// The answer is: we can't just clean up by deleting everything; we have
// to clean up by ReleaseNotifiering everything, and then surrender our pointers.
//
// (2) The only reliable notification we get of a form being deleted from the
// project, is when its module gets a Destroyed message. But the module also
// gets Destroyed when the form is merely being closed. How to distinguish
// between the two cases? After receiving a destroy, we wait a tick, and then
// check whether the form is still contained in its project.
//
// (3) When a module is opened, and it has a form attached, we can get its
// FormEditor, and can create a working notifier from it. But! Maybe the
// source code for the module is open in the IDE, but not yet the form itself.
// In this case, any attempt to GetRootComponent results in an exception.
// How can we tell when a form has actually been opened, so as to check
// its components and send out waExists messages as necessary? There is
// no such notification. Instead, we will implement our own by polling it
// periodically.
//
// (4) I wanted to be able to use _di_TNotIDE &c. so has to have automatic
// reference counting on my objects. It effects the following magic:
// * Whenever you declare a _di_TNot it automatically initializes
// the pointer to NULL
// * Whenever you assign to it a value, it automatically Releases its old
// and AddRefs the new.
// * Whenever you assign to it another thing, it automatically QueryInterfaces.
// For example, _di_TNot n = iotanotifier will work only if iotanotifier happened
// to be a TNot, and will give NULL otherwise. It's similar to dynamic_cast
// !!EXCEPT!! that it gives an access violation of iotanotifier was NULL.
// * Whenever it goes out of scope, it automatically Releases its value.
// * You can use it as a class member as well.
// * You can use DynamicArray<_di_TNot> to make a dynamic array. Again,
// when this dynamic-array goes out of scope (or is a member in a class
// that is being destructed) then every element inside the array will
// also get its own little destructor called (thus Releasing everything)
// For the automatic QueryInterfacing to work I had to supply GUIDs to
// my own classes.
//
// (5) There's no explicit indication of whether a form-that-we're-opening
// was created fresh, or already existed. I deduce this fact as follows:
// if the module filename (ie. unit.cpp) already existed on disk, then the
// form is being opened; if it didn't, then the form is being created.
// (We have to keep track of this value until such time as the form
// actually opens)

class TTheWatcher; TTheWatcher *Watcher=NULL; bool __fastcall EnsureWatcher();
typedef void __fastcall (__closure *TNotEvent)(_di_IOTANotifier);



const GUID IID_TNotIDE = {0x457e1924, 0xECFA, 0x4f32, {0xad, 0xa6, 0x9, 0x99, 0x3d, 0xd3, 0x14, 0x59}};
class TNotIDE; typedef System::DelphiInterface<TNotIDE> _di_TNotIDE;
typedef void __fastcall (__closure *TNotIDEModuleEvent)(_di_IOTANotifier,AnsiString fn);
class INTERFACE_UUID("{457E1924-ECFA-4f32-ADA6-09993DD31459}") TNotIDE : public IOTAIDENotifier50
{
public:
  // IUnknown, IOTANotifier, IOTAIDENotifier, IOTAIDENotifier50
  HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void **ppv) {if (riid==IID_IUnknown || riid==IID_IOTANotifier || riid==IID_IOTAIDENotifier || riid==IID_IOTAIDENotifier50 || riid==IID_TNotIDE) {*ppv=this; AddRef(); return S_OK;} else return E_NOINTERFACE;}
  ULONG STDMETHODCALLTYPE AddRef() {ref_add(this); return InterlockedIncrement(&ref);}
  ULONG STDMETHODCALLTYPE Release() {int tmp = InterlockedDecrement(&ref); ref_dec(this); if (tmp==0) delete this; return tmp;}
  void __fastcall AfterSave(){} void __fastcall BeforeSave(){} void __fastcall Modified(){}
  void __fastcall BeforeCompile(const _di_IOTAProject Project, bool &cancel){}  void __fastcall AfterCompile(bool succeeded){}
  void __fastcall BeforeCompile(const _di_IOTAProject Project, bool isinsight, bool &cancel){}  void __fastcall AfterCompile(bool succeeded,bool isinsight){}
  //
  void __fastcall Destroyed()
  { dbe("idenot. "+name+dptr(this)+" Destroyed"); // should never happen
  }
  void __fastcall FileNotification(TOTAFileNotification fn,const AnsiString FileName,bool &cancel)
  { if (fn==ofnFileOpened)
    { char *c=new char[FileName.Length()+1]; strcpy(c,FileName.c_str());
      PostMessage(ihwnd,WM_APP,0,(LPARAM)c);
    }
  }
  void __fastcall InternalMessageProc(TMessage &Msg)
  { if (Msg.Msg==WM_APP)
    { char *c=(char*)Msg.LParam; AnsiString FileName=c; delete[] c;
      bool fileexists = (Msg.WParam>0);
      OnModule(this,FileName);
    }
    Msg.Result=DefWindowProc(ihwnd,Msg.Msg,Msg.WParam,Msg.LParam);
  }
  //
  __fastcall TNotIDE(AnsiString aname,TNotIDEModuleEvent AOnModuleOpen) : ref(0), name(aname), num(-1), OnModule(AOnModuleOpen)
  { ihwnd=AllocateHWnd(InternalMessageProc);
  }
  __fastcall ~TNotIDE()
  { if (ihwnd!=NULL) DeallocateHWnd(ihwnd); ihwnd=NULL;
  }
  bool __fastcall Start()
  { if (num!=-1) return false;
    s=BorlandIDEServices; num=s->AddNotifier(this);
    db("idenot. "+name);
    _di_IOTAModuleServices ms=BorlandIDEServices;
    for (int i=0; i<ms->ModuleCount; i++) OnModule(this,ms->Modules[i]->GetFileName());
    return true;
  }
  void __fastcall Stop()
  { if (num!=-1)
    { db("[idenot. "+name+" stopping]");
      s->RemoveNotifier(num);
    }
    num=-1;
  }
protected:
  long ref; AnsiString name; HWND ihwnd;
  TNotIDEModuleEvent OnModule;
  int num; _di_IOTAServices s;
};







const GUID IID_TNotMod = {0x642bc7e9, 0x16ca, 0x464d, {0x82, 0xc6, 0xa4, 0xa8, 0x15, 0xcb, 0x3d, 0x7}};
class TNotMod; typedef System::DelphiInterface<TNotMod> _di_TNotMod;
typedef void __fastcall (__closure *TNotModuleFormEvent)(_di_IOTANotifier,_di_IOTAProject p,_di_IOTAModule m,_di_IOTAFormEditor fe);
class INTERFACE_UUID("{642BC7E9-16CA-464d-82C6-A4A815CB3D07}") TNotMod : public IOTAModuleNotifier
{
public:
  // IUnknown, IOTANotifier, IOTAModuleNotifier
  HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void **ppv) {if (riid==IID_IUnknown || riid==IID_IOTANotifier || riid==IID_IOTAModuleNotifier || riid==IID_TNotMod) {*ppv=this; AddRef(); return S_OK;} else return E_NOINTERFACE;}
  ULONG STDMETHODCALLTYPE AddRef() {ref_add(this); return InterlockedIncrement(&ref);}
  ULONG STDMETHODCALLTYPE Release() {ref_dec(this); int tmp = InterlockedDecrement(&ref); if (tmp==0) delete this; return tmp;}
  void __fastcall AfterSave(){} void __fastcall BeforeSave(){} void __fastcall Modified(){}
  bool __fastcall CheckOverwrite() {return true;}
  //
  void __fastcall ModuleRenamed(const AnsiString NewName)
  { AnsiString fn=ExtractFileName(NewName);
    db("mnot. "+name+"->"+fn); name=fn;
  }
  //
  void __fastcall Destroyed()
  { Stop();
    // See note above on how we detect that a form is being deleted from the project
    AnsiString fo="", fn=m->GetFileName(), pr=p->GetFileName();
    for (int i=0; i<p->GetModuleCount(); i++)
    { if (p->GetModule(i)->GetFileName().LowerCase()==m->GetFileName().LowerCase()) fo=p->GetModule(i)->GetFormName();
    }
    AnsiString s=pr+"@"+fn+"@"+fo; char *c=new char[s.Length()+1]; strcpy(c,s.c_str());
    PostMessage(ihwnd,WM_APP,0,(LPARAM)c);
  }
  void __fastcall InternalMessageProc(TMessage &Msg)
  { if (Msg.Msg==WM_APP)
    { char *c=(char*)Msg.LParam; AnsiString s=AnsiString(c); delete[] c;
      int pos=s.Pos("@"); AnsiString pr=s.SubString(1,pos-1); s=s.SubString(pos+1,s.Length()-pos);
      pos=s.Pos("@"); AnsiString fn=s.SubString(1,pos-1), fo=s.SubString(pos+1,s.Length()-pos);
      _di_IOTAProject p=GetProjectFromFileName(pr); bool stillin=false;
      for (int i=0; p!=NULL && !stillin && i<p->GetModuleCount(); i++)
      { if (fn.LowerCase() == p->GetModule(i)->GetFileName().LowerCase()) stillin=true;
      }
      if (!stillin) OnFormRemove(p,NULL,NULL,NULL,fo,"",waDestroy,wsForm);
      OnDestroy(this); return; // might end up destructing us.
    }
    Msg.Result = DefWindowProc(ihwnd,Msg.Msg,Msg.WParam,Msg.LParam);
  }
  //
  __fastcall TNotMod(AnsiString aname,TNotEvent AOnDestroy,TNotModuleFormEvent AOnForm,TWatchEvent AOnFormRemove) : ref(0), name(aname), num(-1), OnDestroy(AOnDestroy), OnForm(AOnForm), OnFormRemove(AOnFormRemove)
  { ihwnd=AllocateHWnd(InternalMessageProc);
  }
  __fastcall ~TNotMod()
  { if (ihwnd!=NULL) DeallocateHWnd(ihwnd); ihwnd=NULL;
  }
  bool __fastcall Start(AnsiString fn)
  { if (num!=-1) return false;
    _di_IOTAModuleServices ms = BorlandIDEServices;
    m = ms->FindModule(fn); if (m==NULL) return false;
    _di_IOTAFormEditor fe = GetFormEditorForModule(m); if (fe==NULL) return false;
    if (m->GetOwnerCount()==0) p=NULL; else p=m->GetOwner(0);
    db("mnot. "+name);
    num = m->AddNotifier(this);
    OnForm(this,p,m,fe); // NB. note above about GetRootComponent
    return true;
  }
  void __fastcall Stop()
  { if (num!=-1)
    { db("[mnot. "+name+" stopping]");
      m->RemoveNotifier(num);
    }
    num=-1;
  }
protected:
  long ref; AnsiString name; HWND ihwnd;
  int num; _di_IOTAProject p; _di_IOTAModule m;
  TNotEvent OnDestroy;
  TNotModuleFormEvent OnForm;
  TWatchEvent OnFormRemove;
};







const GUID IID_TNotForm = {0x732165fe, 0xc07d, 0x4893, {0x95, 0x14, 0x9, 0x4b, 0x3, 0xdd, 0x40, 0x3}};
class TNotForm; typedef System::DelphiInterface<TNotForm> _di_TNotForm;
class INTERFACE_UUID("{732165FE-C07D-4893-9514-094B03DD4003}") TNotForm : public IOTAFormNotifier
{
public:
  // IUnknown, IOTANotifier, IOTAFormNotifier
  HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void **ppv) {if (riid==IID_IUnknown || riid==IID_IOTANotifier || riid==IID_IOTAFormNotifier || riid==IID_TNotForm) {*ppv=this; AddRef(); return S_OK;} else return E_NOINTERFACE;}
  ULONG STDMETHODCALLTYPE AddRef() {ref_add(this); return InterlockedIncrement(&ref);}
  ULONG STDMETHODCALLTYPE Release() {int tmp = InterlockedDecrement(&ref); ref_dec(this); if (tmp==0) delete this; return tmp;}
  void __fastcall AfterSave(){} void __fastcall BeforeSave(){} void __fastcall Modified(){}
  void __fastcall FormActivated(){}  void __fastcall FormSaving(){}
  //
  void __fastcall Destroyed()
  { Stop();
    OnDestroy(this); // might end up destructing us
  }
  typedef struct {TComponent *c;AnsiString sold;AnsiString snew;} TRenameData;
  void __fastcall ComponentRenamed(void* ComponentHandle,const AnsiString OldName,const AnsiString NewName)
  { CheckFormOpen(true);
    if (OldName=="" && NewName=="") return; // a dummy message that gets sent upon component creation
    // See note above about timing of component messages
    if (NewName!="")
    { TRenameData *rd=new TRenameData; rd->c=(TComponent*)ComponentHandle;
      rd->sold=OldName; rd->snew=NewName; PostMessage(ihwnd,WM_APP,0,(LPARAM)rd);
      return;
    }
    // Was it the form being deleted? (in which case we can't pass form info)
    // or just a component?
    _di_IOTAComponent c = fe->GetComponentFromHandle(ComponentHandle);
    bool isroot =  (ComponentHandle == fe->GetRootComponent()->GetComponentHandle());
    if (isroot) OnFormEvent(p,m,NULL,c,OldName,"",waDestroy,wsForm);
    else OnFormEvent(p,m,fe,c,OldName,"",waDestroy,wsComponent);
  }
  void __fastcall InternalMessageProc(TMessage &Msg)
  { if (Msg.Msg==WM_APP)
    { TRenameData *rd=(TRenameData*)Msg.LParam; TComponent *com=rd->c;
      _di_IOTAComponent c = fe->GetComponentFromHandle(com);
      bool isroot =  (com == fe->GetRootComponent()->GetComponentHandle());
      AnsiString sold=rd->sold, snew=rd->snew; delete rd;
      if (isroot)
      { if (sold=="") OnFormEvent(p,m,fe,c,snew,"",waCreate,wsForm);
        else
        { OnFormEvent(p,m,fe,c,snew,sold,waRename,wsForm);
          if (sold==name) {db("fnot. "+sold+"->"+snew); name=snew;}
        }
      }
      else
      { if (sold=="") OnFormEvent(p,m,fe,c,snew,"",waCreate,wsForm);
        else OnFormEvent(p,m,fe,c,snew,sold,waRename,wsComponent);
      }
    }
    if (Msg.Msg==WM_TIMER)
    { if (CheckFormOpen(true)) KillTimer(ihwnd,1);
    }
    Msg.Result = DefWindowProc(ihwnd,Msg.Msg,Msg.WParam,Msg.LParam);
  }
  //
  __fastcall TNotForm(AnsiString aname,bool aisnew,TNotEvent AOnDestroy,TWatchEvent AOnFormEvent,TWatchEvent AOnExistsEvent) : ref(0), name(aname), num(-1), OnDestroy(AOnDestroy), OnFormEvent(AOnFormEvent), OnExistsEvent(AOnExistsEvent), isformopen(false), isnew(aisnew)
  { ihwnd=AllocateHWnd(InternalMessageProc);
  }
  __fastcall ~TNotForm()
  { if (ihwnd!=NULL) DeallocateHWnd(ihwnd);
  }
  bool __fastcall Start(_di_IOTAProject ap,_di_IOTAModule am,_di_IOTAFormEditor afe)
  { if (num!=-1) return false;
    CheckFormOpen(false);
    db("fnot. "+name);
    p=ap; m=am; fe=afe; num = fe->AddNotifier(this);
    // See note above about polling and GetRootComponent
    if (!isformopen) SetTimer(ihwnd,1,350,NULL);
    return true;
  }
  void __fastcall Stop()
  { if (num!=-1)
    { db("[fnot. "+name+" stopping]");
      fe->RemoveNotifier(num);
    }
    num=-1;
  }
  bool __fastcall CheckFormOpen(bool showmessage)
  { if (isformopen) return true;
    _di_IOTAComponent rc=NULL;
    try {rc=fe->GetRootComponent();} catch (Exception &e) {return false;}
    if (rc==NULL) {dbe(name+" has a null root component!"); return false;}
    AnsiString fo=""; bool res=rc->GetPropValueByName("Name",&fo);
    if (!res) {dbe(name+" failed to get name of form component"); return false;}
    if (showmessage) db("fnot. '"+fo+"' opened, "+name);
    name=fo; isformopen=true;
    if (isnew) OnFormEvent(p,m,fe,rc,name,"",waCreate,wsForm);
    else
    { if (OnExistsEvent!=NULL) RecNotifyExistence(rc,true);
    }
    return true;
  }
  void __fastcall RecNotifyExistence(_di_IOTAComponent c,bool isform)
  { AnsiString s=""; bool res=c->GetPropValueByName("Name",&s); if (!res) {dbe("On form "+name+" failed to get name of a component");return;}
    if (isform) OnExistsEvent(p,m,fe,c,s,"",waExists,wsForm);
    else OnExistsEvent(p,m,fe,c,s,"",waExists,wsComponent);
    for (int i=0; i<c->GetComponentCount(); i++) RecNotifyExistence(c->GetComponent(i),false);
  }
protected:
  long ref; AnsiString name; HWND ihwnd; bool isformopen;
  int num; _di_IOTAProject p; _di_IOTAModule m; _di_IOTAFormEditor fe;
  bool isnew; // was this form created fresh?
  TWatchEvent OnFormEvent, OnExistsEvent; // second may be null (because slow!)
  TNotEvent OnDestroy;
};










const GUID IID_TNotDebug = {0xd084473a, 0x1c5, 0x4c5d, {0xba, 0xa8, 0xfb, 0xc7, 0xdf, 0x65, 0x91, 0xb2}};
class TNotDebug; typedef System::DelphiInterface<TNotDebug> _di_TNotDebug;
class INTERFACE_UUID("{D084473A-01C5-4c5d-BAA8-FBC7DF6591B2}") TNotDebug : public IOTADebuggerNotifier
{
public:
  // IUnknown, IOTANotifier, IOTADebuggerNotifier
  HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void **ppv) {if (riid==IID_IUnknown || riid==IID_IOTANotifier || riid==IID_IOTADebuggerNotifier || riid==IID_TNotDebug) {*ppv=this; AddRef(); return S_OK;} else return E_NOINTERFACE;}
  ULONG STDMETHODCALLTYPE AddRef() {ref_add(this); return InterlockedIncrement(&ref);}
  ULONG STDMETHODCALLTYPE Release() {int tmp = InterlockedDecrement(&ref); ref_dec(this); if (tmp==0) delete this; return tmp;}
  void __fastcall AfterSave(){} void __fastcall BeforeSave(){} void __fastcall Modified(){}
  void __fastcall BreakpointAdded(_di_IOTABreakpoint bp) {}  void __fastcall BreakpointDeleted(_di_IOTABreakpoint bp) {}
  //
  void __fastcall Destroyed()
  { dbe("dnot. "+name+dptr(this)+" Destroyed"); // should never happen
  }
  //
  void __fastcall ProcessCreated(_di_IOTAProcess p)
  { int id=p->GetProcessId(); ids->Add((void*)id);
    if (ids->Count==1)
    { db("dnot. "+name+" Debuging Started");
      OnDebug(NULL,NULL,NULL,NULL,"","",waCreate,wsDebug);
    }
  }
  void __fastcall ProcessDestroyed(_di_IOTAProcess p)
  { int id=p->GetProcessId(); int i=ids->IndexOf((void*)id);
    if (i==-1) {dbe("dnot. "+name+" odd! a process finished that never started.");return;}
    ids->Delete(i);
    if (ids->Count==0)
    { db("dnot. "+name+" Debugging finished");
      OnDebug(NULL,NULL,NULL,NULL,"","",waDestroy,wsDebug);
    }
  }
  //
  __fastcall TNotDebug(AnsiString aname,TWatchEvent AOnDebug) : ref(0), name(aname), num(-1), OnDebug(AOnDebug)
  { ids=new TList();
  }
  __fastcall ~TNotDebug()
  { if (ids!=NULL) delete ids; ids=NULL;
  }
  bool __fastcall Start()
  { if (num!=-1) return false;
    ds=BorlandIDEServices;
    num=ds->AddNotifier(this);
    db("dnot. "+name);
    return true;
  }
  void __fastcall Stop()
  { if (num!=-1)
    { db("[dnot. "+name+" stopping]");
      ds->RemoveNotifier(num);
    }
    num=-1;
  }
protected:
  long ref; AnsiString name;
  TWatchEvent OnDebug;
  int num; _di_IOTADebuggerServices ds;
  TList *ids;
};












typedef struct {TWatchSubject ws; TWatchEvent we;} TWatch;
class TWatches : public TList
{ public:
  __fastcall TWatches() : TList() {}
  __fastcall ~TWatches() {for (int i=0; i<Count; i++) {TWatch *w=WItems[i]; delete w; WItems[i]=NULL;}}
  TWatch* __fastcall GetW(int i) {return (TWatch*)Items[i];}
  void __fastcall SetW(int i,TWatch *w) {Items[i]=w;}
  __property TWatch* WItems[int i] = {read=GetW,write=SetW};
  void __fastcall DeleteAllMatchingSubject(TComponent *com)
  { TWatches *todel=new TWatches();
    for (int i=0; i<Count; i++)
    { bool delthis=true;
      TComponent *cw=ExtractWatchEventSubject(WItems[i]->we);
      if (com!=NULL && com!=cw) delthis=false;
      if (delthis) todel->Add(WItems[i]);
    }
    while (todel->Count>0)
    { TWatch *w=todel->WItems[0];
      delete w; // delete the watch
      todel->Delete(0); // remove it from the todel list
      Remove(w); // remove it from the actual list
    }
    delete todel;
  }
};



// TTheWatcher -- has two halves. On one half it sets up and maintains
// IOTANotifiers for the IDE, for every live module, and for every
// life form. On the other half, it maintains a list of clients
// who have asked for TWatchEvents to be sent to them. The middle
// bit receives events from the notifiers, and forwards the events
// on to the relevant clients.
class TTheWatcher : public TComponent
{
public:
  DynamicArray<_di_IOTANotifier> notifiers;
  TWatches *watches;

  __fastcall TTheWatcher(TComponent *Owner) : TComponent(Owner)
  { watches=new TWatches();
  }
  __fastcall ~TTheWatcher()
  { Watcher=NULL; // so that we don't try anything fishy during destruction
    if (watches!=NULL) delete watches; watches=NULL;
    // notifiers is removed by magic
    // NB. we assume that Stop has already been called
  }
  void __fastcall Stop()
  { StopNotifiers(); StopWatches(); db("[w. stopping]");
  }



  // Notifier half. If a thing is in our list of notifiers, then
  // we must eventually stop it.
  void __fastcall Start()
  { db("w.");
    _di_TNotIDE inot = new TNotIDE("theide",&OnModule);
    if (inot->Start()) list(inot);
    _di_TNotDebug dnot = new TNotDebug("thedebugger",&Notify);
    if (dnot->Start()) list(dnot);
  }
  void __fastcall StopNotifiers()
  { for (int i=0; i<notifiers.Length; i++)
    { if (notifiers[i]==NULL) continue;
      _di_TNotIDE inot = notifiers[i]; if (inot!=NULL) inot->Stop();
      _di_TNotMod mnot = notifiers[i]; if (mnot!=NULL) mnot->Stop();
      _di_TNotForm fnot= notifiers[i]; if (fnot!=NULL) fnot->Stop();
      _di_TNotDebug dnot=notifiers[i]; if (dnot!=NULL) dnot->Stop();
    }
  }
  void __fastcall list(_di_IOTANotifier not)
  { for (int i=0; i<notifiers.Length; i++)
    { if (notifiers[i]==NULL) {notifiers[i]=not; return;}
    }
    notifiers.Length ++;
    notifiers[notifiers.Length-1] = not;
  }
  void __fastcall unlist(_di_IOTANotifier not)
  { for (int i=0; i<notifiers.Length; i++)
    { if (notifiers[i]==not)
      { notifiers[i]=NULL;
        return;
      }
    }
    dbe("w. Asking to unlist the unlisted "+dptr(not));
  }
  void __fastcall OnModule(_di_IOTANotifier not,AnsiString fn)
  { _di_TNotMod mnot = new TNotMod(ExtractFileName(fn),&unlist,&OnForm,&Notify);
    if (mnot->Start(fn)) list(mnot);
  }
  void __fastcall OnForm(_di_IOTANotifier not,_di_IOTAProject p,_di_IOTAModule m,_di_IOTAFormEditor fe)
  { bool isnew = !FileExists(m->GetFileName());
    _di_TNotForm fnot = new TNotForm("form of "+ExtractFileName(m->GetFileName()),isnew,&unlist,&Notify,&Notify);
    if (fnot->Start(p,m,fe)) list(fnot);
  }





  // Client half. We keep track of all clients who have
  // registered with us.
  void __fastcall WatchNotification(TWatchSubject ws,TWatchEvent we)
  { TComponent *com=ExtractWatchEventSubject(we);
    com->FreeNotification(this);
    TWatch *w=new TWatch; w->ws=ws; w->we=we; watches->Add(w);
    db("w. watching "+dws(ws)+" for client "+com->Name+" ("+watches->Count+" watches)");
  }
  void __fastcall Notification(TComponent *com,TOperation op)
  { if (op==opRemove && com!=NULL && watches!=NULL)
    { watches->DeleteAllMatchingSubject(com);
      db("w. Stopwatching for client "+com->Name+" ("+watches->Count+" watches left)");
    }
    TComponent::Notification(com,op);
  }
  void __fastcall StopWatches()
  { watches->DeleteAllMatchingSubject(NULL);
  }



  // Middle. Here we get events, and we forward them on to any client
  // who is interested
  void __fastcall Notify(_di_IOTAProject p,_di_IOTAModule m,_di_IOTAFormEditor fe,_di_IOTAComponent c,AnsiString name,AnsiString oldname,TWatchAction wa,TWatchSubject ws)
  { try
    { for (int i=0; i<watches->Count; i++)
      { TWatch *w=watches->WItems[i];
        if (w->ws==ws) w->we(p,m,fe,c,name,oldname,wa,ws);
      }
    } catch (Exception &e) {dbee("w. Sending a notify",e);}
  }

};



bool __fastcall EnsureWatcher()
{ if (Watcher!=NULL) return true;
  try {Watcher=new TTheWatcher(NULL); Watcher->Start(); return true;}
  catch (Exception &e) {dbee("Initializing The Watcher",e); return false;}
}
void __fastcall WatchNotification(TWatchSubject ws,TWatchEvent we)
{ if (!EnsureWatcher()) return;
  try {Watcher->WatchNotification(ws,we);}
  catch (Exception &e) {dbee("WatchNotification ",e);}
}






// VerbPlonker -- special kind of notification. Whenever a component of a
// particular type is dropped on a form, we execute one of its verbs.
// This has to be delayed! Because the OTA still requires a pointer
// to the component for a bit after it has been plonked. So we have
// an internal message.
class TVerbPlonkWatcher : public TComponent
{
public:
  AnsiString type; AnsiString verb; HWND ihwnd;
  __fastcall TVerbPlonkWatcher(TComponent *Owner,AnsiString at,AnsiString av) : TComponent(Owner), type(at), verb(av)
  { SafeSetName(Owner,this,"plnk_"+type);
    ihwnd=AllocateHWnd(InternalMessageProc);
    db("vpw. "+Name+" : "+verb);
  }
  __fastcall ~TVerbPlonkWatcher()
  { if (ihwnd!=NULL) DeallocateHWnd(ihwnd); ihwnd=NULL;
    db("[vpw. "+Name+" stopping]");
  }
  //
  void __fastcall OnPlonk(_di_IOTAProject p,_di_IOTAModule m,_di_IOTAFormEditor fe,_di_IOTAComponent c,AnsiString name,AnsiString oldname,TWatchAction wa,TWatchSubject ws)
  { if (wa!=waCreate || ws!=wsComponent) return;
    AnsiString t=c->GetComponentType(); if (t!=type) return;
    //
    AnsiString cn=GetComponentNameOTA(fe,c);
    AnsiString un=fe->GetModule()->GetFileName();
    AnsiString s=cn+"@"+un;
    char *cc=new char[s.Length()+1]; strcpy(cc,s.c_str());
    PostMessage(ihwnd,WM_APP,0,(LPARAM)cc);
  }
  void __fastcall InternalMessageProc(TMessage &Msg)
  { if (Msg.Msg==WM_APP)
    { char *cc=(char*)Msg.LParam; AnsiString s=AnsiString(cc); delete[] cc;
      int p=s.Pos("@"); AnsiString cn=s.SubString(1,p-1), un=s.SubString(p+1,s.Length()-p);
      _di_IOTAModule m=GetModuleFromFileName(un); if (m==NULL) {dbe("plnk. Failed to get module from unitname "+un);return;}
      _di_IOTAFormEditor fe=GetFormEditorForModule(m); if (fe==NULL) {dbe("plnk. Failed to get form editor for module "+un);return;}
      _di_IOTAComponent c=fe->FindComponent(cn); if (c==NULL) {dbe("plnk. Failed to find component "+cn+" in module "+un);return;}
      ExecuteComponentVerbOTA(fe,c,verb);
    }
    Msg.Result = DefWindowProc(ihwnd,Msg.Msg,Msg.WParam,Msg.LParam);
  }
};
void __fastcall RegisterPlonkVerb(AnsiString type,AnsiString verb)
{ if (!EnsureWatcher()) return;
  TVerbPlonkWatcher *vpw=new TVerbPlonkWatcher((TComponent*)Watcher,type,verb);
  WatchNotification(wsComponent,vpw->OnPlonk);
}




// TNameWatcher -- special kind of watch, for forms or components
// being renamed. A client has asked for us, and given us his OnName event.
// We are created as a child of that client. Therefore, when the
// client is destroyed, we as well will be destroyed.
// Each client may have several OnName events.
// For each client.OnName we have record of which project to watch
// That can be changed by nn(<project>fn,OnName) or nn(<project>0x08,OnName)
// For each client.OnName we have a record of which name to watch
// A watch can be removed with nn("",OnName). (But that still leaves the
// project record behind.) The name "*" will match all forms, and "*.*"
// will match all components.
//
// sform.scom is the thing we're watching. (scom is empty for a form)
//
class TNameWatcher : public TComponent
{ public: TWatchEvent we; _di_IOTAProject p; AnsiString sform,scom;
  //
  __fastcall TNameWatcher(TComponent *Owner,TWatchEvent awe) : TComponent(Owner), p(NULL), we(awe)
  { SafeSetName(Owner,this,"nw_"+Owner->Name);
    db("nw. "+Name);
  }
  __fastcall ~TNameWatcher()
  { p=NULL;
    db("[nw. "+Name+" stopped]");
  }
  //
  static TNameWatcher* __fastcall Ensure(TWatchEvent we)
  { TComponent *Owner=ExtractWatchEventSubject(we);
    for (int i=0; i<Owner->ComponentCount; i++)
    { TNameWatcher *nw=dynamic_cast<TNameWatcher*>(Owner->Components[i]);
      if (nw!=NULL && nw->we==we) return nw;
    }
    TNameWatcher *nw= new TNameWatcher(Owner,we);
    WatchNotification(wsComponent,nw->ComponentEvent);
    WatchNotification(wsForm,nw->FormEvent);
    return nw;
  }
  //
  // This table says precisely how we respond to notifications.
  // fc->f'c' msg means "we used to be searching for sform,scom.
  // now we will search for sform',scom'. also send 'msg' to the client".
  // The top half deals with form events, the bottom with component-events.
  // The left half deals with renaming, the right with destruction.
  //
  //         f c -> f'c' msg              f c -> f'c'  msg
  //
  // fA->B   * *    * *             fAx   * *    * *
  //         * C    * C                   * C    * C
  //         *      *    A->B             *      *     A->
  //         A *    B *  A.*->B.*         A *          A.*->
  //         A C    B C  A.C->B.C         A C          A.C->
  //         A      B    A->B             A            A->
  //         D *    D *                   D *    D *
  //         D C    D C                   D C    D C
  //         D      D                     D      D
  //
  // cF,A->B * *    * *  F.A->F.B   cF,Ax * *    * *   F.A->
  //         * A    * A  F.A->F.B         * A    * A   F.A->
  //         * C    * C                   * C    * C
  //         *      *                     *      *
  //         F *    F *  F.A->F.B         F *    F *   F.A->
  //         F A    F B  F.A->F.B         F A          F.A->
  //         F C    F C                   F C    F C
  //         F      F                     F      F
  //         G *    G *                   G *    G *
  //         G A    G A                   G A    G A
  //         G C    G C                   G C    G C
  //         G      G                     G      G

  void __fastcall ComponentRename(AnsiString F,AnsiString B,AnsiString A)
  { if (sform=="*" && scom=="*") Notify("*","*",F+"."+B,F+"."+A);
    else if (sform=="*" && scom==A) Notify("*",A,F+"."+B,F+"."+A);
    else if (sform==F && scom=="*") Notify(F,"*",F+"."+B,F+"."+A);
    else if (sform==F && scom==A) Notify(F,B,F+"."+B,F+"."+A);
  }
  void __fastcall ComponentDestroy(AnsiString F,AnsiString A)
  { if (sform=="*" && scom=="*") Notify("*","*","",F+"."+A);
    else if (sform=="*" && scom==A) Notify("*",A,"",F+"."+A);
    else if (sform==F && scom=="*") Notify(F,"*","",F+"."+A);
    else if (sform==F && scom==A) Notify("","","",F+"."+A);
  }

  void __fastcall FormRename(AnsiString B,AnsiString A)
  { if (sform=="*" && scom=="") Notify("*","",B,A);
    else if (sform==A)
    { if (scom=="*") Notify(B,"*",B+".*",A+".*");
      else if (scom=="") Notify(B,"",B,A);
      else /*scom==C*/ Notify(B,scom,B+"."+scom,A+"."+scom);
    }
  }
  void __fastcall FormDestroy(AnsiString A)
  { if (sform=="*" && scom=="") Notify("*","","",A);
    else if (sform==A)
    { if (scom=="*") Notify("","","",A+".*");
      else if (scom=="") Notify("","","",A);
      else /*scom==C*/ Notify("","","",A+"."+scom);
    }
  }


  void __fastcall FormEvent(_di_IOTAProject ap,_di_IOTAModule m,_di_IOTAFormEditor fe,_di_IOTAComponent c,AnsiString name,AnsiString oldname,TWatchAction wa,TWatchSubject ws)
  { if (p!=NULL && p!=ap) return;
    tp=p; tm=m; tfe=fe; tc=c; twa=wa; tws=ws;
    if (wa==waRename) FormRename(name,oldname);
    if (wa==waDestroy) FormDestroy(name);
    tp=NULL; tm=NULL; tfe=NULL; tc=NULL;
  }
  _di_IOTAProject tp; _di_IOTAModule tm; _di_IOTAFormEditor tfe; _di_IOTAComponent tc; TWatchAction twa; TWatchSubject tws;
  void __fastcall ComponentEvent(_di_IOTAProject ap,_di_IOTAModule m,_di_IOTAFormEditor fe,_di_IOTAComponent c,AnsiString name,AnsiString oldname,TWatchAction wa,TWatchSubject ws)
  { if (p!=NULL && p!=ap) return;
    if (wa!=waRename && wa!=waDestroy) return;
    _di_IOTAComponent rc = fe->GetRootComponent();
    AnsiString fo=""; bool res=rc->GetPropValueByName("Name",&fo);
    if (!res) {dbe("nw. Failed to get form name "+ExtractFileName(m->GetFileName()));return;}
    tp=p; tm=m; tfe=fe; tc=c; twa=wa; tws=ws;
    if (wa==waRename) ComponentRename(fo,name,oldname);
    if (wa==waDestroy) ComponentDestroy(fo,name);
    tp=NULL; tm=NULL; tfe=NULL; tc=NULL;
  }
  void __fastcall Notify(AnsiString asform,AnsiString ascom,AnsiString name,AnsiString oldname)
  { sform=asform; scom=ascom;
    we(tp,tm,tfe,tc,name,oldname,twa,tws);
  }

};

void __fastcall NameNotification(AnsiString name,TWatchEvent we)
{
  if (name.SubString(1,9)=="<project>") // nn("<project>",ne). To indicate which project to look for notifications
  { _di_IOTAProject p=NULL;
    AnsiString fn=name.SubString(10,name.Length()-9);
    if (fn.SubString(1,2)=="0x")
    { TComponent *com=(TComponent*)(fn.ToInt());
      if (com!=NULL)
      { _di_IOTAModule m=GetModuleContainingComponent(com); p=NULL;
        if (m==NULL) {dbe("w. NameNotification specified component "+com->Name+" is not on any modules");}
        if (m->GetOwnerCount()!=0) p=m->GetOwner(0);
      }
    }
    else if (fn!="")
    { p=GetProjectFromFileName(fn);
      if (p==NULL) {dbe("NameNotification. Failed to find project "+fn);return;}
    }
    TNameWatcher *nw= TNameWatcher::Ensure(we); nw->p=p;
  }

  else // nn("form[->com]",OnName).
  { int p=name.Pos("."); bool isform = (p==0);
    AnsiString sform,scom;
    TNameWatcher *nw = TNameWatcher::Ensure(we);
    if (isform) {nw->sform=name; nw->scom="";}
    else {nw->sform=name.SubString(1,p-1); nw->scom=name.SubString(p+1,name.Length()-p);}
  }

}







// TPropertyExistsWatcher -- special kind of watch, to be notified whenever
// a property of a certain type becomes live. Whenever we get an
// OnExistence notification, we enumerate all its properties.
class TPropertyExistsWatcher : public TComponent
{
public:
  TMetaClass* comclass;
  PTypeInfo proptype;
  AnsiString propname;
  TPropertyWatcherEvent peev;
  __fastcall TPropertyExistsWatcher(TComponent *Owner,PTypeInfo aproptype,TMetaClass* acomclass,AnsiString apropname,TPropertyWatcherEvent apeev) : TComponent(Owner), proptype(aproptype), comclass(acomclass), propname(apropname), peev(apeev)
  { AnsiString n="p"; if (comclass!=NULL) n=n+comclass->ClassName();
    if (propname!=NULL) n=n+propname;
    SafeSetName(Owner,this,n);
    db("propwatch. "+Name);
  }                        
  __fastcall ~TPropertyExistsWatcher()
  { db("[propwatch. "+Name+" stopped.]");
  }
  //
  void __fastcall OnExists(_di_IOTAProject,_di_IOTAModule m,_di_IOTAFormEditor fe,_di_IOTAComponent c,AnsiString name,AnsiString,TWatchAction wa,TWatchSubject ws)
  { if (wa!=waExists && wa!=waCreate) return;
    if (comclass!=NULL && c->GetComponentType()!=comclass->ClassName()) return;
    // now enum through the properties. We'll have to get some interfaces first.
    _di_INTAFormEditor nfe=NULL;
    HRESULT res=fe->QueryInterface(__uuidof(INTAFormEditor),(void**)&nfe); if (FAILED(res)) return;
    _di_IFormDesigner ides=nfe->GetFormDesigner();
    _di_IPersistent ip=c->GetIPersistent();
    ip->AddRef(); TPersistent *p=TryExtractPersistent(ip); // manual reference counting
    TDesignerSelectionList *list=new TDesignerSelectionList(); list->Add(p);
    TTypeKinds kinds; kinds << tkUnknown << tkInteger << tkChar << tkEnumeration << tkFloat
      << tkString << tkSet << tkClass << tkMethod << tkWChar << tkLString << tkWString
      << tkVariant << tkArray << tkRecord << tkInterface;
    GetComponentProperties(list,kinds,ides,EnumProc);
    delete list;
    ip->Release();
  }
  void __fastcall EnumProc(TPropertyEditor *pe)
  { bool send=true;
    if (proptype!=NULL && proptype!=pe->GetPropType()) send=false;
    if (propname!="" && propname!=pe->GetName()) send=false;
    if (!send) return;
    if (pe->PropCount==0) return;
    db("propwatch. Found a property "+pe->GetName()+" - notifying...");
    if (peev!=NULL) peev(pe);
  }
};

void __fastcall RegisterPropertyWatcher(PTypeInfo proptype,TMetaClass* comclass,AnsiString propname,TPropertyWatcherEvent peev)
{ if (!EnsureWatcher()) return;
  TPropertyExistsWatcher *pew=new TPropertyExistsWatcher((TComponent*)Watcher,proptype,comclass,propname,peev);
  WatchNotification(wsComponent,pew->OnExists);
  WatchNotification(wsForm,pew->OnExists);
}











class TTheWatcherLifetimeExpert : public IOTAWizard
{
public:
  // IUnknown, IOTANotifier
  HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void **ppv) {if (riid==IID_IUnknown || riid==IID_IOTANotifier || riid==IID_IOTAWizard) {*ppv=this; AddRef(); return S_OK;} else return E_NOINTERFACE;}
  ULONG STDMETHODCALLTYPE AddRef() {ref_add(this); return InterlockedIncrement(&ref);}
  ULONG STDMETHODCALLTYPE Release() {ref_dec(this); int tmp = InterlockedDecrement(&ref); if (tmp==0) delete this; return tmp;}
  void __fastcall AfterSave(){}  void __fastcall BeforeSave(){}  void __fastcall Destroyed(){}  void __fastcall Modified(){}
  // IOTAWizard.
  AnsiString __fastcall GetIDString() {return watcherid;}
  AnsiString __fastcall GetName() {return "Lifetime Of The Watcher";}
  TWizardState __fastcall GetState() {TWizardState ws; ws << wsEnabled; return ws;}
  void __fastcall Execute() {}
  // TTheWatcherLifetimeExpert
  __fastcall TTheWatcherLifetimeExpert() : IOTAWizard(), ref(0)
  { db(watcherid);
  }
  __fastcall ~TTheWatcherLifetimeExpert()
  { if (Watcher!=NULL)
    { try {Watcher->Stop(); delete Watcher; Watcher=NULL;}
      catch (Exception &e) {dbee("w. Destroying the watcher",e);}
    }
    db("["+watcherid+" stopped]");
  }
private:
  long ref;
};















namespace Toolsapiplus
{ void __fastcall PACKAGE Register()
  { RegisterPackageWizard(new TTheWatcherLifetimeExpert());
  }
}



