#ifndef toolsapiplusH
#define toolsapiplusH
#include <toolsapi.hpp>



// INSTRUCTIONS ON USING TOOLSAPIPLUS
// If you want to use toolsapiplus in your own package,
// you MUST RENAME IT! That's because BCB doesn't allow two
// packages both containing a unit toolsapiplus. So, you
// should call yours mypkg_toolsapiplus.cpp or something like that.






// IOTA NAVIGATION FUNCTIONS
// Getting from one useful interface to another, conveniently
// The final call GetFormEditorEx will open the form if it is not already
// open.
_di_IOTAProject __fastcall GetCurrentProject();
_di_IOTAProjectResource __fastcall GetProjectResources(_di_IOTAProject p);
_di_IOTAModule __fastcall GetModuleContainingComponent(TComponent *c);
_di_IOTAProject __fastcall GetProjectFromFileName(AnsiString fn);
_di_IOTAProject __fastcall GetModuleAsProject(_di_IOTAModule m);
_di_IOTAFormEditor __fastcall GetFormEditorForModule(_di_IOTAModule m);
_di_IDesigner __fastcall GetDesignerFromFormEditor(_di_IOTAFormEditor fe);
_di_IOTAFormEditor GetFormEditorEx(AnsiString fo,AnsiString un);
_di_IOTAModule __fastcall GetModuleFromFileName(AnsiString un);
_di_IOTAModuleServices __fastcall GetModuleServices();
_di_IOTAProject __fastcall GetProjectFromDesigner(_di_IFormDesigner ides);
void __fastcall EnsureUniqueModuleName(AnsiString *ounit,AnsiString *oform,AnsiString *odir);
bool __fastcall LookForComponent(_di_IOTAProject p,AnsiString *name,AnsiString *type, AnsiString *form,AnsiString *unit);



// RESOURCE FUNCTIONS
// These will add resources to the project. You must save the project module yourself
// after adding them. AddIconResource returns the ICONGROUP resource, not an
// individual icon resource. For all of them you can either pass a
// ProjectResource argument, or you can leave it NULL (in which case it
// uses the current default project). Both Add functions assume that they
// are given a file (either an icon file or a bitmap file, either on
// disk with filename fn, or an RT_RCDATA resource image of that file with *RESNAME
_di_IOTAResourceEntry __fastcall AddIconFileResource(AnsiString fn, char *name,WORD flags,WORD langid,int DataVersion,int Version,int Characteristics,_di_IOTAProjectResource pr);
_di_IOTAResourceEntry __fastcall AddBitmapFileResource(AnsiString fn,char *name,WORD flags,WORD langid,int DataVersion,int Version,int Characteristics,_di_IOTAProjectResource pr);
bool __fastcall DeleteIconGroup(char *name,_di_IOTAProjectResource pr);




// PROPERTY ACCESS FUNCTIONS
// These functions access the properties through their property editors.
// The effect is just like typing something in at the object inspector.
AnsiString __fastcall GetComponentPropertyTextOTA(_di_IOTAFormEditor fe,_di_IOTAComponent c,AnsiString prop);
AnsiString __fastcall GetComponentPropertyText(_di_IFormDesigner ides,_di_IPersistent ip,AnsiString prop);
void __fastcall SetComponentPropertyTextOTA(_di_IOTAFormEditor fe,_di_IOTAComponent c,AnsiString prop,AnsiString val);
void __fastcall SetComponentPropertyText(_di_IFormDesigner ides,_di_IPersistent ip,AnsiString prop,AnsiString val);
AnsiString __fastcall GetComponentNameOTA(_di_IOTAFormEditor fe,_di_IOTAComponent c);
AnsiString __fastcall GetComponentName(_di_IFormDesigner ides,_di_IPersistent ip);
void __fastcall SetComponentImageOTA(_di_IOTAFormEditor fe,_di_IOTAComponent c,AnsiString fn,bool isbmp);
void __fastcall SetComponentImage(_di_IFormDesigner ides,_di_IPersistent ip,AnsiString fn,bool isbmp);
bool __fastcall ExecuteComponentVerb(_di_IFormDesigner ides,_di_IPersistent ip,AnsiString verb);



// TSERVICESFILE
// This is a convenient(!) way to do IOTAFiles. Create an object of this
// type, passing a string in the constructor. You can then use it as
// an IOTAFile interface, for anything else that needs one.
class TServicesFile : public IOTAFile
{
private:
  AnsiString s; TDateTime dt; long ref;
public:
  TServicesFile(AnsiString as) : IOTAFile(), s(as), dt(-1), ref(0) {}
  // IUnknown
  HRESULT STDMETHODCALLTYPE QueryInterface(REFIID riid, void **ppv) {if (riid==IID_IUnknown || riid==IID_IOTAFile) {*ppv=this; AddRef(); return S_OK;} else return E_NOINTERFACE;}
  ULONG STDMETHODCALLTYPE AddRef() {return InterlockedIncrement(&ref);}
  ULONG STDMETHODCALLTYPE Release() {int tmp = InterlockedDecrement(&ref); if (tmp==0) delete this; return tmp;}
  // IOTAFile
  AnsiString __fastcall GetSource() {return s;}
  TDateTime __fastcall GetAge() {return dt;}
};








// TFORMSCANNER. For scanning through all forms and components in a
// project. It provides events for scanning modules, forms, components, so
// you can customize it do to pretty much anything. Easiest way to use
// it is just to call Scan(getproc,blank,forms,components,hiddens)
// which generates a list and also calls getproc on each entry in the list.
// The boolean values say whether to include the name of each form,
// include the name of each component, include hidden forms (TForm1...)
// which are in the project but have not yet been opened in the IDE.
typedef void __fastcall (__closure *TScanProjectEvent)(AnsiString fn,bool &cancel);
typedef void __fastcall (__closure *TScanModuleEvent)(AnsiString fn,AnsiString un,AnsiString fo,AnsiString dc,bool &cancel);
typedef void __fastcall (__closure *TScanFormEvent)(TCustomForm *form,AnsiString &s,bool &cancel);
typedef void __fastcall (__closure *TScanComponentEvent)(TCustomForm *form,TComponent *com,AnsiString &s,bool &cancel);
typedef void __fastcall (__closure *TScanHiddenFormEvent)(AnsiString fo,AnsiString un,AnsiString &s,bool &cancel);
class TFormScanner : public TObject
{
public:
  __fastcall TFormScanner();
  __fastcall ~TFormScanner();
  virtual void __fastcall Scan();
  virtual void __fastcall QuickScan(_di_IFormDesigner ides,TGetStrProc proc,bool af,bool ac,bool ah);
  virtual bool __fastcall StringStandsForHiddenForm(AnsiString s);
  virtual bool __fastcall OpenFormIfHidden(AnsiString val);
  //
  TScanProjectEvent OnScanProject;
  TScanHiddenFormEvent OnScanHiddenForm;
  TScanModuleEvent OnScanModule;
  TScanFormEvent OnScanForm;
  TScanComponentEvent OnScanComponent;
  TScanComponentEvent OnScanSubComponent;
  TStringList *results; TStringList *res2; virtual void __fastcall Add(AnsiString caption,TObject *obj,AnsiString s);
  bool autoforms,autocomponents,autohiddens,dosubcomponents; _di_IOTAProject autoproject;
  TGetStrProc GetStrProc;
protected:
  TStringList *hiddens;
  virtual void __fastcall ScanProject(_di_IOTAProject p);
  virtual void __fastcall ScanModule(_di_IOTAModuleInfo mi);
  virtual void __fastcall ScanForm(_di_IOTAFormEditor fe);
  virtual void __fastcall ScanComponent(_di_IOTAFormEditor fe,_di_IOTAComponent c,int depth,AnsiString path);
};






// WATCH.
// These typedefs are used in conjunction with the Watch notifiers
// wtExists is if the thing already existed. wsDebug refers to a debugging
// session that gets either started (waCreate) or stopped (waDestroy).
enum TWatchSubject {wsForm, wsComponent, wsDebug};
enum TWatchAction {waCreate,waDestroy,waExists,waRename};

// WatchNotification -- to handle the various sorts of watch events
// that might come. Must be an event in a TComponent descendent.
// the 'name' parameter is the subject of the message; the optional 'oldname'
// is only used for renaming.
// There are several specialised sort of WatchNotifications down below...
typedef void __fastcall (__closure *TWatchEvent)(_di_IOTAProject p,_di_IOTAModule m,_di_IOTAFormEditor fe,_di_IOTAComponent c,AnsiString name,AnsiString oldname,TWatchAction wa,TWatchSubject ws);
void __fastcall WatchNotification(TWatchSubject ws,TWatchEvent we);

// RegisterPlonk -- so that some action is taken as soon as a particular
// component type is dropped somewhere
//   RegisterPlonkVerb("TSwappingComponent","Swap...");
void __fastcall RegisterPlonkVerb(AnsiString type,AnsiString verb);

// PropertyExistenceNotification -- creates one instance of watcherclass for
// for every live instance of the matching property. watcherclass must
// descend from TPropertyWatcher.
typedef void __fastcall (*TPropertyWatcherEvent)(TPropertyEditor *pe);
void __fastcall RegisterPropertyWatcher(PTypeInfo proptype,TMetaClass* comclass,AnsiString propname,TPropertyWatcherEvent peev);

// NameNotification -- notify whenever "Name" gets renamed (or deleted)
//    nn("form",OnName) or nn("form->component",OnName)
//    nn("<project>c:\prj1.bpk",OnName) or nn("<project>0x004800>",OnName)
// the last call means we'll only be triggered by renames in that project
// Stop watch with nn("",OnName)
// This TWatchEvent argument has a slightly different semantics from before.
// Before, name was always the subject and oldname only applied to renaming.
// But here, name is always the new value and oldname always the old.
// So deletion would be name='',oldname='something'
void __fastcall NameNotification(AnsiString n,TWatchEvent we);








// TPropertyWatcher -- base class for property watchers. For each instance
// of the corresponding property that exists at designtime, there will be
// exactly one property-watcher. You must RegisterPropertyWatcher on some
// static method in your property editor. That static method should create
// an instance of this.
// OnExists should be called immediately after construction.
// OnSet should be called after the property editor has done SetValue
class TPropertyWatcher : public TComponent
{
public:
  virtual __fastcall TPropertyWatcher(TComponent *Owner); AnsiString propname;
  virtual void __fastcall OnExists(TPropertyEditor *pe);
  virtual void __fastcall OnSet(TPropertyEditor *pe,AnsiString sold,AnsiString snew);
};



// TFarPropertyWatcher -- If you have a property editor contains the text
// name of another form, and that other form gets renamed, then your property
// value needs to be updated automatically. First, put an OnExists
// static method in your property editor to create a FarPropertyWatcher.
// (the ComponentNameEditor, FormNameEditor and FormTypeEditor already have
// such a method). Second, as well as doing RegisterPropertyEditor, you
// also have to call RegisterPropertyWatcher(myprop::OnExists). See the
// example component in sample.cpp to see how its done.
// The FarPropertyWatcher watches for changes to a value:
// * It is given an initial value upon streaming in. We get this
// in the TPropertyNotifier constructor. We must start watching for this.
// * Its value is changed when the user types something in,
// so calling SetValue. We must stop watching the old, and start watching
// the new.
// * We must change its value upon OnName
class TFarPropertyWatcher : public TPropertyWatcher
{
public:
  __fastcall TFarPropertyWatcher(TComponent *Owner); bool ignorethis; AnsiString prefix;
  virtual void __fastcall Watch(AnsiString s);
  AnsiString __fastcall Decorate(AnsiString s); AnsiString __fastcall Undecorate(AnsiString s);
  void __fastcall OnExists(TPropertyEditor *pe);
  void __fastcall OnSet(TPropertyEditor *pe,AnsiString sold,AnsiString snew);
  void __fastcall OnName(_di_IOTAProject p,_di_IOTAModule m,_di_IOTAFormEditor fe,_di_IOTAComponent c,AnsiString name,AnsiString oldname,TWatchAction wa,TWatchSubject ws);
};




// TFormScanningEditor -- an abstract class for the sort of properties
// that are names to something else in the project
class TFormScanningEditor : public TPropertyEditor
{
public:
  virtual __fastcall TFormScanningEditor(const _di_IFormDesigner ides,int count); AnsiString prefix;
  virtual __fastcall ~TFormScanningEditor();
  virtual void __fastcall GenerateList()=NULL; TFormScanner *scanner;
  //
  void __fastcall GetValues(TGetStrProc Proc);
  TPropertyAttributes __fastcall GetAttributes();
  AnsiString __fastcall GetValue();
  void __fastcall SetValue(AnsiString NewValue);
  //
  virtual void __fastcall StoreValue(AnsiString v)=NULL;
  virtual AnsiString __fastcall RetrieveValue()=NULL;
};



// TComponetnNameEditor -- a property that is a name of a component on a distant form.
// If using it, you must also RegisterPropertyWatcher(TFarComponentEditor::OnExists)
class TComponentNameEditor : public TFormScanningEditor
{
public:
  virtual __fastcall TComponentNameEditor(const _di_IFormDesigner ides,int count);
  void __fastcall GenerateList();
  static void __fastcall OnExists(TPropertyEditor *pe);
};


// TFormNameEditor -- an property that is a name of some other form in the project.
// If using it, you must also RegisterPropertyWatcher(TFarFormEditor::OnExists)
class TFormNameEditor : public TFormScanningEditor
{
public:
  virtual __fastcall TFormNameEditor(const _di_IFormDesigner ides,int count);
  void __fastcall GenerateList();
  void __fastcall OnScanModule(AnsiString fn,AnsiString un,AnsiString fo,AnsiString dc,bool &cancel);
  static void __fastcall OnExists(TPropertyEditor *pe);
};


// TFormTypeEditor -- an property that is a name of some other form in the project.
// If using it, you must also RegisterPropertyWatcher(TFormTypeEditor::OnExists)
// NOTE - A thing that's useful is to have a property with form-types,
// and then dynamically create an instance of that class at runtime. For
// instance, have a design-time property called 'ErrorForm' with
// value TMyErrForm1. Then at runtime, whenever an error occurs, you
// dynamically create a TMyErrForm1 to display upon.
// For this to work you need for TMyErrForm1 to have been RegisterClass'd
// at runtime. The easiest way to do this is for the user of your component
// to #include "watchingsysdefs.h" near the top of her main project source
// code, after #pragma hdrstop, and before USEFORMs. That will
// RegisterClass all her forms automatically. Or she can do it manually
// if she wants.
class TFormTypeEditor : public TFormScanningEditor
{
public:
  virtual __fastcall TFormTypeEditor(const _di_IFormDesigner ides,int count);
  void __fastcall GenerateList();
  void __fastcall OnScanModule(AnsiString fn,AnsiString un,AnsiString fo,AnsiString dc,bool &cancel);
  AnsiString __fastcall Decorate(AnsiString s);
  AnsiString __fastcall Undecorate(AnsiString s);
  static void __fastcall OnExists(TPropertyEditor *pe);
};








// TSwappingComponentEditor -- a component-editor with a single
// verb: "swap". When the user chooses this it displays a popup
// menu containing a list of types. The user selects one, and
// the component is replaced by one of that other type.
// Override GetSwapText() to call it something else
// Override ExecuteSwapMenu() to populate the menu.
// At the end of ExecuteSwapMenu, you should ExecuteSwap(ExecuteMenu())
class TSwappingComponentEditor : public TComponentEditor
{
public:
  virtual __fastcall TSwappingComponentEditor(TComponent* Component, _di_IFormDesigner Designer);
  virtual __fastcall ~TSwappingComponentEditor();
  int __fastcall GetVerbCount();
  AnsiString __fastcall GetVerb(int i);
  void __fastcall Edit();
  void __fastcall ExecuteVerb(int i);
protected:
  HMENU hmenu; TStringList *names, *types;
  virtual int __fastcall AddMenuItem(AnsiString s,AnsiString n);
  virtual int __fastcall ExecuteMenu();
  virtual _di_IOTAComponent __fastcall ExecuteSwap(int i);
  virtual _di_IOTAComponent __fastcall CreateComponent(_di_IOTAFormEditor fe,_di_IOTAComponent par,AnsiString type,AnsiString name,int x,int y,int w,int h);
  // override these two:
  virtual AnsiString __fastcall GetSwapText();
  virtual void __fastcall ExecuteSwapMenu() = NULL;
};





#endif
